#include <string.h>

#include "command.h"
#include "constants.h"

#define STARTSIGN '#'
#define ENDSIGN '\r'

void addStart ( char* string )
{
    int length = strlen ( string );
    string[length] = STARTSIGN;
}

void addEnd ( char* string )
{
    int length = strlen ( string );
    string[length] = ENDSIGN;
}

void addAddress ( char* string, int address )
{
    int length = strlen ( string );
    string[length] = address + 'a';
}

void addCommand ( char* string, char command )
{
    int length = strlen ( string );
    string[length] = command;
}

void addData ( char* string, char* data )
{
    unsigned int length = strlen ( string );
    unsigned int i = 0;
    unsigned dataLength = strlen ( data );

    for ( i = 0; i<dataLength; i++ )
    {
        string[length+i] = data[i];
    }
}

void encode64 ( char data[150], char* encoded, unsigned int length )
{
    unsigned int pt = 0;
    unsigned char a,b,c;
    unsigned char ptr = 0;

    while ( length > 0 )
    {
        if ( length )
        {
            a = data[ptr++];
            length--;
        }
        else a = 0;
        if ( length )
        {
            b = data[ptr++];
            length--;
        }
        else b = 0;
        if ( length )
        {
            c = data[ptr++];
            length--;
        }
        else c = 0;

        encoded[pt++] = '=' + ( a >> 2 );
        encoded[pt++] = '=' + ( ( ( a & 0x03 ) << 4 ) | ( ( b & 0xf0 ) >> 4 ) );
        encoded[pt++] = '=' + ( ( ( b & 0x0f ) << 2 ) | ( ( c & 0xc0 ) >> 6 ) );
        encoded[pt++] = '=' + ( c & 0x3f );
    }
    encoded[pt] = 0;
}

void decode64 ( char *encodedData, int length, char *decodedData )
{
    unsigned char a,b,c,d;
    unsigned char ptr = 0;
    unsigned char x,y,z;

    int offset = 0;

    if ( encodedData[offset] == 0 )
    {
        return;
    }

    while ( length != 0 )
    {
        a = encodedData[offset++] - '=';
        b = encodedData[offset++] - '=';
        c = encodedData[offset++] - '=';
        d = encodedData[offset++] - '=';

        x = ( a << 2 ) | ( b >> 4 );
        y = ( ( b & 0x0f ) << 4 ) | ( c >> 2 );
        z = ( ( c & 0x03 ) << 6 ) | d;

        if ( length-- ) decodedData[ptr++] = x;
        else break;
        if ( length-- ) decodedData[ptr++] = y;
        else break;
        if ( length-- ) decodedData[ptr++] = z;
        else break;
    }
}

void addCRC ( char* string )
{
    unsigned int tmpCRC = 0;
    char CRC[2];
    int i = 0;
    int length = strlen ( string );

    for ( i = 0; i < strlen ( string ); i++ )
    {
        tmpCRC += string[i];
    }
    tmpCRC %= 4096;

    CRC[0] = '=' + tmpCRC / 64;
    CRC[1] = '=' + tmpCRC % 64;
    CRC[2] = '\0';

    string[length+0] = CRC[0];
    string[length+1] = CRC[1];
    string[length+2] = CRC[2];
}

int checkCRC ( char *encodedData, int length )
{
    int crc = 0;
    int i = 0;

    if ( length < 4 ) return FAILURE;

    if ( encodedData[1] == 127 )
        encodedData[1] = 0;

    for ( i=0; i < length-2; i++ )
        crc+=encodedData[i];

    crc = crc % 4096;

    if ( encodedData[length - 2] != ( '=' + ( crc / 64 ) ) )
        return FAILURE;

    if ( encodedData[length - 1] != ( '=' + crc % 64 ) )
        return FAILURE;

    return SUCCESS;
}
