#include <math.h>
#include <stdio.h>
#include "types.h"
#include "gps.h"

double rad2deg(double radian) {
    return radian * 180 / M_PI;
}

double deg2rad(double degree) {
    return degree * M_PI / 180;
}

double rad2km(double radian, double radius) {
    return radius * radian;
}

double km2rad(double km, double radius) {
    return km / radius;
}
int isInTargetRange(GPS_Pos_t position, GPS_Pos_t target) {
    int isInRange = FALSE;
    
    double lat_deg_from = ((double) position.Latitude / 10000000.0);
    double lon_deg_from = ((double) position.Longitude / 10000000.0);
    double lat_deg_to = ((double) target.Latitude / 10000000.0);
    double lon_deg_to = ((double) target.Longitude / 10000000.0);
    
    double radius_km = 6371.0;
    double dLat = deg2rad(lat_deg_to - lat_deg_from);
    double dLon = deg2rad(lon_deg_to - lon_deg_from);
    double lat_rad_from = deg2rad(lat_deg_from);
    double lat_rad_to = deg2rad(lat_deg_to);
    double a = sin(dLat/2.0) * sin(dLat/2.0) + sin(dLon/2.0) * sin(dLon/2.0) * cos(lat_rad_from) * cos(lat_rad_to);
    double c = 2.0 * atan2(sqrt(a), sqrt(1.0-a));
    double d = radius_km * c;
    
    if(d*1000.0 <= IN_RANGE_RADIUS) {
      isInRange = TRUE;
    }
    
    return isInRange;
}

void calcGPSPosition(double lat_deg, double lon_deg, double distance_km, double angle_deg, double* newLat, double* newLon) {
    double lat_rad = deg2rad(lat_deg);
    double lon_rad = deg2rad(lon_deg);
    double newLat_rad = 0.0;
    double newLon_rad = 0.0;
    double radius_km = 6371.0;
    double distance_rad = km2rad(distance_km, radius_km);
    double angle_rad = deg2rad(-angle_deg); //clockwise

    newLat_rad = asin(sin(lat_rad) * cos(distance_rad) + cos(lat_rad) * sin(distance_rad) * cos(angle_rad));
    if (cos(newLat_rad) == 0) //pole
        newLon_rad = lon_rad;
    else
        newLon_rad = fmod(lon_rad - asin(sin(angle_rad) * sin(distance_rad) / cos(newLat_rad)) + M_PI, 2 * M_PI) - M_PI;

    *newLat = rad2deg(newLat_rad);
    *newLon = rad2deg(newLon_rad);
}

double calcBearing(double lat_deg_from, double lon_deg_from, double lat_deg_to, double lon_deg_to) {
    double lat_rad_from = deg2rad(lat_deg_from);
    double lon_rad_from = deg2rad(lon_deg_from);
    double lat_rad_to = deg2rad(lat_deg_to);
    double lon_rad_to = deg2rad(lon_deg_to);
    double dLon = deg2rad(lon_deg_from - lon_deg_to);
    double y = sin(dLon) * cos(lat_rad_to);
    double x = cos(lat_rad_from) * sin(lat_rad_to) - sin(lat_rad_from) * cos(lat_rad_to) * cos(dLon);
    double bearing = rad2deg(atan2(y,x));
    
    if(bearing < 0 ) {
      bearing = bearing * -1;
    } else if(bearing > 0) {
      bearing = 360-bearing;
    }
  
    return bearing;
}

double negateBearing(double bearing_orig) {
    double bearing = 0;
    bearing = bearing_orig + 180.0;
    if(bearing >= 360.0) {
      bearing -= 360.0;
    }
    return bearing;
}
  
void calcSpiralWaypoints(GPS_Pos_t * spiralWaypoints, GPS_Pos_t position, double r, double incr_deg, int start_deg, int max_turns) {
    GPS_Pos_t tmpPosition;
    tmpPosition.Altitude = position.Altitude;
    tmpPosition.Status = position.Status;
    
    printf("int lat: %d, lon: %d\n", position.Latitude, position.Longitude);

    double lat_deg = ((double) position.Latitude / 10000000);
    double lon_deg = ((double) position.Longitude / 10000000);
    printf("lat_deg: %f, lon_deg: %f\n", lat_deg, lon_deg);
    double spiral_start_lat; //p1
    double spiral_start_lon; //p1
    double distance = r;
    double spiral_tmp_lat;
    double spiral_tmp_lon;
    double incr_dist = r / 360 * incr_deg;
    printf("r: %f, incr_deg: %f, incr_dist: %f\n", r, incr_deg, incr_dist);
    int waypointCounter = 0;

    int turn = 0;
    int deg = 0;

    calcGPSPosition(lat_deg, lon_deg, r, start_deg, &spiral_start_lat, &spiral_start_lon);
    tmpPosition.Latitude = (int32_t) (spiral_start_lat * 10000000);
    tmpPosition.Longitude = (int32_t) (spiral_start_lon * 10000000);
    spiralWaypoints[waypointCounter++] = tmpPosition;
    
    for (turn = 0; turn < max_turns; turn++) {
        for (deg = start_deg + incr_deg; deg <= 360; deg += incr_deg) {
            distance += incr_dist;
            calcGPSPosition(lat_deg, lon_deg, distance, deg, &spiral_tmp_lat, &spiral_tmp_lon);
            tmpPosition.Latitude = (int32_t) (spiral_tmp_lat * 10000000);
            tmpPosition.Longitude = (int32_t) (spiral_tmp_lon * 10000000);
            spiralWaypoints[waypointCounter++] = tmpPosition;
        }
    }
}
