#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h> 
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>


#include <ifaddrs.h>

#include "types.h"
#include "helper.h"
#include "constants.h"

const int endian_const = 1;
#define is_bigendian() ( (*(char*)&endian_const) == 0 )

void resetInputString(char *inputString, int length, int *inputStringCounter) {
    memset(inputString, '\0', length);
}

uint32_t endian_swap_u32(uint32_t x) {
  if (is_bigendian()) {
	x = (x>>24) | 
	    ((x<<8) & 0x00FF0000) |
	    ((x>>8) & 0x0000FF00) |
	    (x<<24);
  }
  return x;
}

int32_t endian_swap_32(int32_t x) {
  if (is_bigendian()) {
	x = ((x << 8) & 0xFF00FF00) | ((x >> 8) & 0xFF00FF ); 
	x = (x << 16) | ((x >> 16) & 0xFFFF);
  }
  return x;
}

uint16_t endian_swap_u16(uint16_t x) {
  if (is_bigendian()) {
	x = (x>>8) | 
	    (x<<8);
  }
  return x;
}

int16_t endian_swap_16(int16_t x) {
  if (is_bigendian()) {
	x = (x << 8) |
	   ((x >> 8) & 0xFF);
  }
  return x;
}

void endian_swap_navidata(NaviData_t* navidata) {
  navidata->CurrentPosition.Longitude = endian_swap_32(navidata->CurrentPosition.Longitude);
  navidata->CurrentPosition.Latitude = endian_swap_32(navidata->CurrentPosition.Latitude);
  navidata->CurrentPosition.Altitude = endian_swap_32(navidata->CurrentPosition.Altitude);
  
  navidata->TargetPosition.Longitude = endian_swap_32(navidata->TargetPosition.Longitude);
  navidata->TargetPosition.Latitude = endian_swap_32(navidata->TargetPosition.Latitude);
  navidata->TargetPosition.Altitude = endian_swap_32(navidata->TargetPosition.Altitude);
  
  navidata->TargetPositionDeviation.Distance = endian_swap_u16(navidata->TargetPositionDeviation.Distance);
  navidata->TargetPositionDeviation.Bearing = endian_swap_16(navidata->TargetPositionDeviation.Bearing);
  
  navidata->HomePosition.Longitude = endian_swap_32(navidata->HomePosition.Longitude);
  navidata->HomePosition.Latitude = endian_swap_32(navidata->HomePosition.Latitude);
  navidata->HomePosition.Altitude = endian_swap_32(navidata->HomePosition.Altitude);
  
  navidata->HomePositionDeviation.Distance = endian_swap_u16(navidata->HomePositionDeviation.Distance);
  navidata->HomePositionDeviation.Bearing = endian_swap_16(navidata->HomePositionDeviation.Bearing);
  
  navidata->Altimeter = endian_swap_16(navidata->Altimeter);
  navidata->Variometer = endian_swap_16(navidata->Variometer);
  navidata->FlyingTime = endian_swap_u16(navidata->FlyingTime);
  navidata->GroundSpeed = endian_swap_u16(navidata->GroundSpeed);
  navidata->Heading = endian_swap_16(navidata->Heading);
  navidata->CompassHeading = endian_swap_16(navidata->CompassHeading);
  navidata->TopSpeed = endian_swap_16(navidata->TopSpeed);
  navidata->SetpointAltitude = endian_swap_16(navidata->SetpointAltitude);
  navidata->Current = endian_swap_u16(navidata->Current);
  navidata->UsedCapacity = endian_swap_u16(navidata->UsedCapacity);
}

void endian_swap_notification(Notification_t* notification) {
  notification->Position.Longitude = endian_swap_32(notification->Position.Longitude);
  notification->Position.Latitude = endian_swap_32(notification->Position.Latitude);
  notification->Position.Altitude = endian_swap_32(notification->Position.Altitude);
  
  notification->Height = endian_swap_16(notification->Height);
  notification->Heading = endian_swap_16(notification->Heading);
  notification->Speed = endian_swap_u16(notification->Speed);
  endian_swap_status(&(notification->Status));
}

void endian_swap_status(Status_t* status) {
  status->Client1.Position.Longitude = endian_swap_32(status->Client1.Position.Longitude);
  status->Client1.Position.Latitude = endian_swap_32(status->Client1.Position.Latitude);
  status->Client1.Position.Altitude = endian_swap_32(status->Client1.Position.Altitude);
  status->Client2.Position.Longitude = endian_swap_32(status->Client2.Position.Longitude);
  status->Client2.Position.Latitude = endian_swap_32(status->Client2.Position.Latitude);
  status->Client2.Position.Altitude = endian_swap_32(status->Client2.Position.Altitude);
  status->Position.Longitude = endian_swap_32(status->Position.Longitude);
  status->Position.Latitude = endian_swap_32(status->Position.Latitude);
  status->Position.Altitude = endian_swap_32(status->Position.Altitude);
}

int newMAC(uint8_t newMAC[6], uint8_t oldMAC[6]) {
    int i = 0;
    int differentMACs = FALSE;
    for (i = 0; i < 6; i++) {
        if (newMAC[i] != oldMAC[i]) {
            differentMACs = TRUE;
            break;
        }
    }
    return differentMACs;

}

void outputIp(struct sockaddr_storage * ip) {
    char str[INET6_ADDRSTRLEN];
    if(ntohs(ip->ss_family) == AF_INET) {
        printf("ipv4: ");
        inet_ntop(AF_INET, &((struct sockaddr_in *) (ip))->sin_addr, str, INET6_ADDRSTRLEN);
    } else if(ntohs(ip->ss_family) == AF_INET6) {
        printf("ipv6: ");
        inet_ntop(AF_INET6, &((struct sockaddr_in6 *) (ip))->sin6_addr, str, INET6_ADDRSTRLEN);
    }
    printf("addr: %s\n", str);
}

void getIp(struct sockaddr_storage * ip, char * ipStr) {
    if(ntohs(ip->ss_family) == AF_INET) {
        printf("ipv4: ");
        inet_ntop(AF_INET, &((struct sockaddr_in *) (ip))->sin_addr, ipStr, INET6_ADDRSTRLEN);
    } else if(ntohs(ip->ss_family) == AF_INET6) {
        printf("ipv6: ");
        inet_ntop(AF_INET6, &((struct sockaddr_in6 *) (ip))->sin6_addr, ipStr, INET6_ADDRSTRLEN);
    }
}

void getMAC(uint8_t* mac, char * interface) {
    struct ifreq ifr;
    int sock;
    int i;

    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (-1 == sock) {
        perror("get mac socket() ");
    }

    strncpy(ifr.ifr_name, interface, sizeof (ifr.ifr_name) - 1);
    ifr.ifr_name[sizeof (ifr.ifr_name) - 1] = '\0';

    if (-1 == ioctl(sock, SIOCGIFHWADDR, &ifr)) {
        perror("get mac ioctl(SIOCGIFHWADDR) ");
    }
    for (i = 0; i < 6; i++) {
        mac[i] = (unsigned char) ifr.ifr_hwaddr.sa_data[i];
    }
}

int getIfn(char * interface) {
    int ifn = 0;

    while (ifn == 0) {
        printf("iface: %s\n", interface);

        ifn = if_nametoindex(interface);

        if (ifn == 0) {
            printf("Interface %s not available (yet) - sleeping\n", (char*) interface);
            sleep(1);
        } else {
            break;
        }
    }
    return ifn;
}
