#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <termios.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include "types.h"
#include "constants.h"

#include "../uavext/uavext.h"
#include "../uavint/uavint.h"

#include "../common_files/helper.h"
#include "../common_files/gps.h"

#include "pinger.h"
#include "notifier.h"
#include "pinglistener.h"
#include "serialreader.h"
#include "simulator.h"
#include "controller.h"


int serialfd, tty;
int readingdebug = 0;
int waypointcounter = 0;
int serialConnectionReady = FALSE;
int mkInitialized = FALSE;
int pingReceived = FALSE;
int stateFixCounter = -1;
int maxSignalStrength = -1000;
int maxSignalCounter = 0;
int aborted = FALSE;
long oldC1 = 0;
int sameSignalStrengthCounter = 0;
int airborneSignalStrengthCounter = 0;

FILE *input;
FILE *output;

struct termios oldkey, newkey; //place for old and new port settings for keyboard teletype
struct termios oldtio, newtio; //place for old and new port settings for serial port
struct sigaction saio; //definition of signal action

char * serialdevice;
char meshinterface[80];
char wlaninterface[80];
char bridgeinterface[80];

int newsockfd;
Status_t currentStatus;
pthread_mutex_t currentStatusMutex = PTHREAD_MUTEX_INITIALIZER;
NaviData_t currentNaviData;
pthread_mutex_t currentNaviDataMutex = PTHREAD_MUTEX_INITIALIZER;
AllowedClients_t currentAllowedClients;
pthread_mutex_t currentAllowedClientsMutex = PTHREAD_MUTEX_INITIALIZER;

Status_t uavs[MAX_UAVS];
int uavscounter = 0;
pthread_mutex_t uavsMutex = PTHREAD_MUTEX_INITIALIZER;

struct sockaddr_storage notificationSubscriber[MAX_NOTIFICATION_SUBSCRIBER];
int notificationSubscriberCounter = 0;
int startedSimulation = 0;
int simulationStep = 0;
int simulation = 0;

Submitted_Pos_t submittedPositions[3];
int submittedPositionsCounter = 0;
int maxSubmittedPositions = 2;

Submitted_Pos_t flightDirectionPosition;
volatile int listeningForSerialInput = TRUE;

char macstring[20];

NaviData_t navidata;

int main(int argc, char *argv[]) {
  
    getConfigVariables();

    fd_set master; // master file descriptor list
    fd_set read_fds; // temp file descriptor list for select()
    fd_set write_fds; // temp file descriptor list for select()
    int fdmax = 0; // maximum file descriptor number

    int listener; // listening socket descriptor
    int newfd; // newly accept()ed socket descriptor
    struct sockaddr_storage remoteaddr; // client address
    socklen_t addrlen;
    
    Status_t lastUav;
    memset(&lastUav, '\0', sizeof(Status_t));

    struct thread_data{
      int  *fd;
      char *device;
    };

    int i;

    struct timeval tv;

    tv.tv_sec = 2;
    tv.tv_usec = 0;

    FD_ZERO(&master); // clear the master and temp sets
    FD_ZERO(&read_fds);
    FD_ZERO(&write_fds);

    memset(notificationSubscriber, 0, sizeof (struct sockaddr) * MAX_NOTIFICATION_SUBSCRIBER);
    memset(uavs, 0, sizeof(Status_t) * MAX_UAVS);

    if (checkInputs(argc, argv) == FAILURE) {
        usage(argv[0]);
    }

    if (initStatus() == FAILURE) {
        printf("Could not initialize the status, exiting...\n");
        exit(EXIT_FAILURE);
    }
    
    pthread_t pinger, notifier, serialreader, pinglistener;
    pthread_create(&pinger, NULL, ping, bridgeinterface);
    pthread_create(&notifier, NULL, notify, bridgeinterface);
    pthread_create(&pinglistener, NULL, listen_for_pings, bridgeinterface);
    
    struct thread_data thread_data_array[1];
    thread_data_array[0].fd = &serialfd;
    thread_data_array[0].device = serialdevice;
    
    pthread_create(&serialreader, NULL, readserial, (void *) &thread_data_array);

    listener = generateControllerCommunicationSocket();

    memset(&saio, 0, sizeof (struct sigaction));
    char tmpSerialBuffer[MAXLINELENGTH];
    char serialBuffer[MAXLINELENGTH];
    int res;
    int loop = 1; /* loop while TRUE */
    memset(tmpSerialBuffer, '\0', sizeof (tmpSerialBuffer));
    memset(serialBuffer, '\0', sizeof (serialBuffer));
    int buf_pointer = 0;
    struct timeval serialTimeOut;
    serialTimeOut.tv_sec = 0;
    serialTimeOut.tv_usec = 1000000;
    
    FD_SET(listener, &master);
    fdmax = listener;
    int socketCounter = 0;
    int multihopFinalPositionCounter = 0;
	
    
    //main loop
    for (;;) {
      
        read_fds = master;
	int n;
	n = select(fdmax + 1, &read_fds, NULL, NULL, &tv);
        if (n == -1) {
            printf("error: select\n");
            exit(4);
        }

        socketCounter = 0;
	
	// run through the existing connections looking for data to read
        for (i = 0; i <= fdmax; i++) {
            if (FD_ISSET(i, &read_fds)) {
		if (i == listener) {
		    printf("i == listener\n");
                    addrlen = sizeof remoteaddr;
                    socketCounter++;
                    newfd = acceptControllerSocket(listener); // handle new connections

                    if (newfd == -1) {
                        perror("accept");
                    } else {
                        FD_SET(newfd, &master);
                        if (newfd > fdmax) {
                            fdmax = newfd;
                        }
                    }
                } else if (i != serialfd) {
                    printf("i != serialfd\n");
                    socketCounter++;

                    // handle data from a client
                    if (handlingCommunication(i) <= 0) {
			printf("closing socket %d\n", i);
                        close(i);
                        FD_CLR(i, &master);
                    }
                } else if (i == serialfd) {
		  printf("i == serialfd\n");
                }
            }
        }
	
	handleUavs();

	if (serialConnectionReady == FALSE && serialfd >= 0) {
            int n = 0;
            n = outSerialLinkTest(serialfd, SERIALCONNECTIONPATTERN);             
        }
	
	int currentState = getCurrentStatusState();
	int currentSearching = getCurrentStatusSearching();
	printf("                                                                                                                                         current State: %d\n", currentState);
	    
	if (serialConnectionReady == TRUE) {
            if (mkInitialized == FALSE) {
		initMk(serialfd);
            }
	    
            n = outSerialDebugInterval(serialfd, 1000); //interval in 10ms
	    
            outSerialOSDInterval(serialfd, 20);
	    navidata = getCurrentNaviData();
	    printf("current lat: %ld, lon: %ld # ", navidata.CurrentPosition.Latitude, navidata.CurrentPosition.Longitude);
	    printf("target lat: %ld, lon: %ld # ", navidata.TargetPosition.Latitude, navidata.TargetPosition.Longitude);
	    printf("home lat: %ld, lon: %ld # ", navidata.HomePosition.Latitude, navidata.HomePosition.Longitude);
	    printf("#WPs: %d, currentWP: %d\n", navidata.WaypointNumber, navidata.WaypointIndex);
	    
	    int currentPositioning = getCurrentStatusPositioning();
	    
	    if(currentState == UAV_STATE_AIRBORNE_FLYING_CENTER) {
		checkForFinalPosition(1, UAV_STATE_AIRBORNE_REACHED_CENTER);
	    }
	    
	    if(currentState == UAV_STATE_AIRBORNE_AWAITING_POS2) {
	      if(currentSearching == UAV_SEARCHING_MANUAL) {
		  checkForFinalPosition(1, UAV_STATE_AIRBORNE_REACHED_MAX_DISTANCE);
	      } else if(currentSearching == UAV_SEARCHING_AUTONOMOUS) {
		  checkForFinalPosition(13, UAV_STATE_AIRBORNE_REACHED_MAX_DISTANCE);
	      }
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_1_AWAITING_POS2) {
	      if(currentSearching == UAV_SEARCHING_MANUAL) {
		  checkForFinalPosition(1, UAV_STATE_MULTIHOP_1_REACHED_MAX_DISTANCE);
	      } else if(currentSearching == UAV_SEARCHING_AUTONOMOUS) {
		  checkForFinalPosition(13, UAV_STATE_MULTIHOP_1_REACHED_MAX_DISTANCE);
	      }
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_1_FLYING_CENTER) {
		if(checkForFinalPosition(1, UAV_STATE_MULTIHOP_1_FLYING_FINAL_POSITION) == TRUE) {  
		  outSerialSendWaypoint(getCurrentStatusClient1().Position, TRUE, 1, serialfd);
		  outSerialSendWaypoint(getCurrentStatusClient1().Position, TRUE, 2, serialfd);
		}
	    }
	    
	    
	    if(currentState == UAV_STATE_AIRBORNE_REACHED_CENTER && currentPositioning == UAV_POSITIONING_SIGNAL) {
	        if(airborneSignalStrengthCounter == 0) {
		  long clientssignals[2] = {0, 0};
		  readSignalStrength(wlaninterface, clientssignals);
		  long c1 = clientssignals[0];
		  long c2 = clientssignals[1];
		  long diff = 0;
		  double bearing = 0;
		  double targetLon = 0.0;
		  double targetLat = 0.0;
		  GPS_Pos_t targetPosition;
		  targetPosition.Longitude = 0;
		  targetPosition.Latitude = 0;
		  targetPosition.Altitude = 0;
		  targetPosition.Status = 0;
		  Submitted_Pos_t client2Pos;
		  printf( "final client 1 signal: %d\n", c1);
		  printf( "final client 2 signal: %d\n", c2);
		  
		  if(c1 < 0 && c2 < 0) {
		      diff = c1 - c2;
		      if(abs(diff) > MAXSIGNALSTRENGTHDIFF) {
			  printf("signal strength diff\n");
			  client2Pos = getCurrentStatusClient2();			
			  bearing = calcBearing(navidata.CurrentPosition.Latitude / 10000000.0, navidata.CurrentPosition.Longitude / 10000000.0, client2Pos.Position.Latitude / 10000000.0, client2Pos.Position.Longitude / 10000000.0);
			  if(c1 < c2) { // fly to c1, 
			      printf("flying to c1 ");
			      bearing = negateBearing(bearing);
			  } else { //else fly to c2
			      printf("flying to c2 ");
			  }
			  calcGPSPosition(navidata.CurrentPosition.Latitude / 10000000.0, navidata.CurrentPosition.Longitude / 10000000.0, 0.002, bearing, &targetLat, &targetLon); // 2m
			  targetPosition.Longitude = targetLon * 10000000.0;
			  targetPosition.Latitude = targetLat * 10000000.0;
			  targetPosition.Altitude = 3000;
			  targetPosition.Status = 1;
			  
			  outSerialSendWaypoint(targetPosition, FALSE, 1, serialfd); //delete waypoint list
			  outSerialSendWaypoint(targetPosition, TRUE, 1, serialfd);
		      } else {
			outSerialSendWaypoint(targetPosition, FALSE, 1, serialfd); //delete waypoint list
		      }
		  }
		}
		airborneSignalStrengthCounter++;
		if(airborneSignalStrengthCounter >= 50) {
		  airborneSignalStrengthCounter = 0;
		}
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_1_FLYING_FINAL_POSITION) {
		Submitted_Pos_t client1Pos, client2Pos;
		long clientssignals[2] = {0, 0};
		GPS_Pos_t targetPosition;
		memset(&targetPosition, '\0', sizeof(GPS_Pos_t));
		double bearing = 0;
		double targetLon = 0.0;
		double targetLat = 0.0;
		targetPosition.Longitude = 0;
		targetPosition.Latitude = 0;
		targetPosition.Altitude = 0;
		targetPosition.Status = 0;
		
		readSignalStrength(wlaninterface, clientssignals);
		long c1 = clientssignals[0];
		printf( "final client 1 signal: %d\n", c1);
		if(c1 == 0) {
		  if(multihopFinalPositionCounter < 50) {
		    multihopFinalPositionCounter++;
		    printf("multihopFinalPositionCounter: %d\n", multihopFinalPositionCounter);
		  } else {
		    if(checkForFinalPosition(1, UAV_STATE_MULTIHOP_1_FLYING_FINAL_POSITION_MANUAL) == TRUE) {
		      client1Pos = getCurrentStatusClient1();
		      client2Pos = getCurrentStatusClient2();    
		      bearing = calcBearing(client1Pos.Position.Latitude / 10000000.0, client1Pos.Position.Longitude / 10000000.0, client2Pos.Position.Latitude / 10000000.0, client2Pos.Position.Longitude / 10000000.0);
		      calcGPSPosition(client1Pos.Position.Latitude / 10000000.0, client1Pos.Position.Longitude / 10000000.0, 0.003, bearing, &targetLat, &targetLon); // 4m
		      targetPosition.Longitude = targetLon * 10000000.0;
		      targetPosition.Latitude = targetLat * 10000000.0;
		      targetPosition.Altitude = 3000;
		      targetPosition.Status = 1;
		      outSerialSendWaypoint(targetPosition, TRUE, 1, serialfd);
		      outSerialSendWaypoint(targetPosition, TRUE, 2, serialfd);
		    }
		  }
		} else if(c1 < 0) {
		    if(abs(abs(OPTIMALSIGNALSTRENGTH) - abs(c1)) <= MAXSIGNALSTRENGTHDIFF) {
			outSerialSendWaypoint(targetPosition, FALSE, 1, serialfd);
			setCurrentStatusState(UAV_STATE_MULTIHOP_1_REACHED_FINAL_POSITION);
		    }
		    if(c1 == oldC1) {
			sameSignalStrengthCounter++;
		    } else {
			sameSignalStrengthCounter = 0;
		    }
		    if(sameSignalStrengthCounter > 50) {
			outSerialSendWaypoint(targetPosition, FALSE, 1, serialfd);
			setCurrentStatusState(UAV_STATE_MULTIHOP_1_REACHED_FINAL_POSITION);
			sameSignalStrengthCounter = 0;
		    }		    
		}
		oldC1 = c1;
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_1_FLYING_FINAL_POSITION_MANUAL) {
		checkForFinalPosition(1, UAV_STATE_MULTIHOP_1_REACHED_FINAL_POSITION);
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_2_FLYING_LAST_UAV) {
		checkForFinalPosition(1, UAV_STATE_MULTIHOP_2_REACHED_LAST_UAV);
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_3_FLYING_LAST_UAV) {
		checkForFinalPosition(1, UAV_STATE_MULTIHOP_3_REACHED_LAST_UAV);
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_2_REACHED_LAST_UAV) {
		Submitted_Pos_t client2Pos = getCurrentStatusClient2();
		outSerialSendWaypoint(client2Pos.Position, FALSE, 1, serialfd);
		outSerialSendWaypoint(client2Pos.Position, TRUE, 1, serialfd);
		outSerialSendWaypoint(client2Pos.Position, TRUE, 2, serialfd);
		setCurrentStatusState(UAV_STATE_MULTIHOP_2_FLYING_FINAL_POSITION);
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_3_REACHED_LAST_UAV) {
		Submitted_Pos_t client2Pos = getCurrentStatusClient2();
	        outSerialSendWaypoint(client2Pos.Position, FALSE, 1, serialfd);
		outSerialSendWaypoint(client2Pos.Position, TRUE, 1, serialfd);
		outSerialSendWaypoint(client2Pos.Position, TRUE, 2, serialfd);
		setCurrentStatusState(UAV_STATE_MULTIHOP_3_FLYING_FINAL_POSITION);
	    }
	    
	    if(currentState == UAV_STATE_MULTIHOP_2_FLYING_FINAL_POSITION) {
		Status_t lastUavLocal;
		memset(&lastUavLocal, '\0', sizeof(Status_t));
		long tmpSignalStrengthLastUav = 0;
		long tmpSignalStrengthDiff = 0;
		lastUavLocal = getUAVByState(UAV_STATE_MULTIHOP_1_REACHED_FINAL_POSITION);
		tmpSignalStrengthLastUav = getUavSignalStrength(meshinterface, lastUavLocal.MAC);
		printf("\nlast uav signal strength: %d\n\n", tmpSignalStrengthLastUav);
		tmpSignalStrengthDiff = abs(abs(OPTIMALMESHSIGNALSTRENGTH) - abs(tmpSignalStrengthLastUav));
		printf("signal diff: %d\n", tmpSignalStrengthDiff);
		    
		if(tmpSignalStrengthDiff <= MAXSIGNALSTRENGTHDIFF) {
			outSerialSendWaypoint(navidata.CurrentPosition, FALSE, 1, serialfd);
			setCurrentStatusState(UAV_STATE_MULTIHOP_2_REACHED_FINAL_POSITION);
		}
	    }
	    if(currentState == UAV_STATE_MULTIHOP_3_FLYING_FINAL_POSITION) {
		Status_t lastUavLocal;
		memset(&lastUavLocal, '\0', sizeof(Status_t));
		long tmpSignalStrengthLastUav = 0;
		long tmpSignalStrengthDiff = 0;
		lastUavLocal = getUAVByState(UAV_STATE_MULTIHOP_2_REACHED_FINAL_POSITION);
		tmpSignalStrengthLastUav = getUavSignalStrength(meshinterface, lastUavLocal.MAC);
		
		printf("\nlast uav signal strength: %d\n\n", tmpSignalStrengthLastUav);
		tmpSignalStrengthDiff = abs(abs(OPTIMALMESHSIGNALSTRENGTH) - abs(tmpSignalStrengthLastUav));
		printf("signal diff: %d\n", tmpSignalStrengthDiff);
		    
		if(tmpSignalStrengthDiff <= MAXSIGNALSTRENGTHDIFF) {
			outSerialSendWaypoint(navidata.CurrentPosition, FALSE, 1, serialfd);
			setCurrentStatusState(UAV_STATE_MULTIHOP_3_REACHED_FINAL_POSITION);
		}
	    }
        }
	usleep(100000);
    }

    return EXIT_SUCCESS;
}

Status_t getUAVByState(int state) {
    Status_t targetUav;
    Status_t tmpUav;
    int i = 0;
    memset(&targetUav, '\0', sizeof(Status_t));
    memset(&tmpUav, '\0', sizeof(Status_t));
    for(i = 0; i<uavscounter; i++) {
	tmpUav = uavs[i];
	if(tmpUav.State == state) {
	  targetUav = tmpUav;
	  return targetUav;
	}
    }
    return targetUav;
}

void handleUavs() {
    int i = 0;
    int multihop1available = 0;
    int multihop2available = 0;
    extern int aborted;
    
    uint8_t currentState = getCurrentStatusState();
    Status_t tmpUav;
    memset(&tmpUav, 0, sizeof(Status_t));
    Status_t lastUav;
    memset(&lastUav, '\0', sizeof(Status_t));
    
        
    pthread_mutex_lock(&uavsMutex);
  
    if(stateFixCounter == 0) {
	if(currentState == UAV_STATE_MULTIHOP_2_WAITING) {
	    setCurrentStatusState(UAV_STATE_MULTIHOP_3_WAITING);
	    stateFixCounter = -1;
	    aborted = FALSE;
	} else if(currentState == UAV_STATE_MULTIHOP_3_WAITING) {
	    setCurrentStatusState(UAV_STATE_MULTIHOP_2_WAITING);
	    aborted = FALSE;
	    
	    stateFixCounter = -1;
	}
    }
    
    for(i = 0; i<uavscounter; i++) {
	tmpUav = uavs[i];
	if(tmpUav.Scenario == UAV_SCENARIO_MULTIHOP) {
	    if(currentState == UAV_STATE_UNCONFIGURED  || currentState == UAV_STATE_ABORTED) {
		if( tmpUav.State == UAV_STATE_MULTIHOP_1_AWAITING_POS1 || 
		    tmpUav.State == UAV_STATE_MULTIHOP_1_AWAITING_FLIGHT_DIRECTION ||
		    tmpUav.State == UAV_STATE_MULTIHOP_1_AWAITING_POS2) {
		    multihop1available = 1;
		    setCurrentStatusConfiguration(tmpUav.Scenario, tmpUav.Searching, tmpUav.Positioning);		    
		}
		if( tmpUav.State == UAV_STATE_MULTIHOP_2_WAITING) {
		    multihop2available = 1;
		}
	    } else if(currentState == UAV_STATE_MULTIHOP_2_WAITING) {
	        if( tmpUav.State == UAV_STATE_MULTIHOP_1_REACHED_FINAL_POSITION) {
		    printf("fly to last uav: %s lat: %ld, lon: %ld \n", tmpUav.Hostname, tmpUav.Position.Latitude, tmpUav.Position.Longitude);
		    outSerialSendWaypoint(tmpUav.Position, FALSE, 1, serialfd);
		    outSerialSendWaypoint(tmpUav.Position, TRUE, 1, serialfd);
		    outSerialSendWaypoint(tmpUav.Position, TRUE, 2, serialfd);
		    setCurrentStatusState(UAV_STATE_MULTIHOP_2_FLYING_LAST_UAV);
		}
	    } else if(currentState == UAV_STATE_MULTIHOP_3_WAITING) {
		if( tmpUav.State == UAV_STATE_MULTIHOP_2_REACHED_FINAL_POSITION) {
		    outSerialSendWaypoint(tmpUav.Position, FALSE, 1, serialfd);
		    outSerialSendWaypoint(tmpUav.Position, TRUE, 1, serialfd);
		    outSerialSendWaypoint(tmpUav.Position, TRUE, 2, serialfd);
		    setCurrentStatusState(UAV_STATE_MULTIHOP_3_FLYING_LAST_UAV);
		}
	    }
	}
    }
    if(multihop1available == 1) {
	if(multihop2available == 1) {
	    setCurrentStatusState(UAV_STATE_MULTIHOP_3_WAITING);
	    aborted = FALSE;	    
	} else {
	    setCurrentStatusState(UAV_STATE_MULTIHOP_2_WAITING);
	    aborted = FALSE;
	}
    }
    pthread_mutex_unlock(&uavsMutex);
}

int checkForFinalPosition(int nbrWaypoints, int newState) {
    int reachedFinalPosition = FALSE;
    NaviData_t navidata;
    GPS_Pos_t position;    
    navidata = getCurrentNaviData();
    if(navidata.WaypointIndex == nbrWaypoints + 1) {
        printf("navidata.WaypointIndex == nbrWaypoints + 1\n");
        memset(&position, 0, sizeof(GPS_Pos_t));
	outSerialSendWaypoint(position, FALSE, 1, serialfd);
	setCurrentStatusState(newState);
	reachedFinalPosition = TRUE;
    }
    return reachedFinalPosition;
}


int getUavSignalStrength(char *interface, uint8_t *mac) {
    char macString[100];
    char command[30];
    char line[300];
    char firstword[100];
    char secondword[100];
    int signalValue = 0;
    int n = 0;
    char *station = "Station";
    char *signal = "signal:";
    
    memset(command, '\0', sizeof(command));
    memset(macString, '\0', sizeof(macString));
    memset(line, '\0', sizeof(line));
    memset(firstword, '\0', sizeof(firstword));
    memset(secondword, '\0', sizeof(secondword));
    
    sprintf(macString, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x",mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
    
    sprintf(command, "iw %s station dump", interface);
    
    FILE *file = popen(command, "r");
    if (file == NULL) {
        printf("cannot open file\n");
	pclose(file) ;
        return signalValue;
        exit(1);
    }
    
    int stationFound = FALSE;
    int signalFound = FALSE;
    
    while (fgets(line, sizeof (line), file) != NULL && signalFound == FALSE) {
	n = sscanf(line, "%s %s", firstword, secondword);
	
	if(stationFound == FALSE) {
	  if (strcmp(firstword, station) == 0 && strcmp(secondword, macString) == 0) {
	    stationFound = TRUE;
	  }
	} else if(stationFound == TRUE) {
	  if (strcmp(firstword, signal) == 0) {
	      signalValue = atol(secondword);
	      signalFound = TRUE;
	  }
	}
    }
    pclose(file) ;
    return signalValue;
}

int readSignalStrength(char *interface, long * clientssignals) {

    char command[30];
    char line[300];
    char firstword[100];
    char secondword[100];
    int lcount = 0;
    int n = 0;
    char *station = "Station";
    char *signal = "signal:";
    int foundstation = TRUE;
    int signalvalue = 1;
    int maxstations = 3;
    char stations[5][20];
    char signals[5][10];
    char clients[2][20];
    char clientsmacstrings[2][20];
    memset(stations, '\0', sizeof(stations));
    memset(signals, '\0', sizeof(signals));
    memset(clients, '\0', sizeof(clients));
    memset(clientsmacstrings, '\0', sizeof(clientsmacstrings));
    int stationscounter = 0;
    int signalscounter = 0;
    int clientscounter = 0;
    int clientssignalcounter = 0;
    sprintf(command, "iw %s station dump", interface);

    FILE *file = popen(command, "r");

    if (file == NULL) {
        printf("cannot open file\n");
	pclose(file);
        return signalvalue;
        exit(1);
    }
    
    getCurrentClient1MacString(clientsmacstrings[0]);
    getCurrentClient2MacString(clientsmacstrings[1]);
    
    while (fgets(line, sizeof (line), file) != NULL && stationscounter <= maxstations) {
        lcount++;
	n = sscanf(line, "%s %s", firstword, secondword);

	if (strcmp(firstword, station) == 0) {
	    strcpy(stations[stationscounter++], secondword);
	}
	
	if (strcmp(firstword, signal) == 0) {
	    strcpy(signals[signalscounter++], secondword);
	}
    }
    int i=0, j=0;
    int tmpsignal = 0;
    for(i = 0; i<maxstations; i++) {
      if (strcmp(stations[i], clientsmacstrings[0]) == 0 || strcmp(stations[i], clientsmacstrings[1]) == 0) {
//	printf("found station %d\n", i);
      }
    }
    
    for(i=0; i<2; i++) {
	tmpsignal = 0;
	for(j=0; j<maxstations; j++) {
	    if(strcmp(clientsmacstrings[i], stations[j]) == 0) {
	      tmpsignal = atol(signals[j]);
	      if(tmpsignal < 0) {
		clientssignals[i] = tmpsignal;
	      } else {
		clientssignals[i] = 0;
	      }
	    }
	}
    }
    pclose(file);

    return signalvalue;
}

int initMk(int serialfd) {

    int n = 0;
    GPS_Pos_t position;
    position.Longitude = 0;
    position.Latitude = 0;
    position.Altitude = 0;
    position.Status = 0;

    printf("mk initializing...\n");
    outSerialSendWaypoint(position, FALSE, 1, serialfd);
    printf("mk initialized, n: %d\n", n);
    mkInitialized = TRUE;
    return n;
}

NaviData_t getCurrentNaviData() {
    NaviData_t navidata;
    pthread_mutex_lock(&currentNaviDataMutex);
    navidata = currentNaviData;
    pthread_mutex_unlock(&currentNaviDataMutex);
    return navidata;
}

void handlingSerialInput(char *inputString, int length) {
    int i = 0;
    uint16_t address;
    uint8_t command;
    char data[MAXDATALENGTH];
    char debugString[length];
    memset(debugString, '\0', sizeof ( debugString));
    memset(data, '\0', sizeof ( data));
    strncpy(debugString, inputString, length - 1);
    if (strlen(inputString) > 5 && checkCRC(inputString, strlen(inputString) - 1)) {
        decodeInternalCommand(inputString, &address, &command, data);
        if (readingdebug == 1) {
            printf("<- %s, address: %d, command: %c\r\n", debugString, address, command);
        }

        switch (command) {
            case 'Z':
                inSerialLinkTest(data);
                break;
            case 'O':
                inSerialOSDData(data);
                break;
            case 'W':
                inNbrOfWaypoints(data);
                break;
            case 'X':
                inWaypoint(data);
                break;
        }
    } else {//too short string or wrong crc
        printf ( "wrong crc received\r\n" );
	for (i = 0; i < strlen(inputString); i++) {
            printf("%c", inputString[i]);
        }
        printf("\r\n");
    }
}

void inNbrOfWaypoints(char *data) // 'W'
{
    uint8_t nbrOfWaypoints = 0;
    memcpy(&nbrOfWaypoints, data, sizeof ( nbrOfWaypoints));
    printf("number of Waypointsreceived: %d\n", nbrOfWaypoints);
}


void inWaypoint(char *data) // 'X'
{
    uint8_t nbrOfWaypoints = 0;
    uint8_t waypointIndex = 0;
    Waypoint_t waypoint;
    memset(&waypoint, '\0', sizeof (Waypoint_t));
    
    printf("inWaypoint ");
    
    memcpy((uint8_t*) & nbrOfWaypoints, data, sizeof ( nbrOfWaypoints));
    data = data + sizeof ( nbrOfWaypoints);
    memcpy((uint8_t*) & waypointIndex, data, sizeof ( waypointIndex));
    data = data + sizeof ( waypointIndex);
    memcpy((Waypoint_t*) & waypoint, data, sizeof ( Waypoint_t));

    waypoint.Position.Longitude = endian_swap_32(waypoint.Position.Longitude);
    waypoint.Position.Latitude = endian_swap_32(waypoint.Position.Latitude);
    waypoint.Position.Altitude = endian_swap_32(waypoint.Position.Altitude);
    
    printf("#WPs: %d, WP Index: %d ", nbrOfWaypoints, waypointIndex);
    printf("lon: %f, lat: %f, altitude: %d, status: %d, heading: %d, tolerance: %d, holdtime: %d, event: %d, index:%d\r\n", waypoint.Position.Longitude / 10000000.0, waypoint.Position.Latitude / 10000000.0, waypoint.Position.Altitude, waypoint.Position.Status, waypoint.Heading, waypoint.ToleranceRadius, waypoint.HoldTime, waypoint.Event_Flag, waypoint.Index);
    
}

void inSerialLinkTest(char *data) // 'z'
{
    uint16_t pattern;

    memcpy((uint16_t*) & pattern, data, sizeof ( uint16_t));
    printf("seriallinkpattern %d\n", pattern);

    if (pattern == SERIALCONNECTIONPATTERN) {
        printf("pattern correct\n");
        serialConnectionReady = TRUE;
    }
}

void inSerialOSDData(char *data) // 'O'
{
    NaviData_t navidata;

    memcpy((NaviData_t*) & navidata, data, sizeof (NaviData_t));
    endian_swap_navidata(&navidata);
    setCurrentNaviData(navidata);
    setCurrentStatusPosition(navidata);

}

int initStatus() {
    int result = FAILURE;
    int i = 0;
    Submitted_Pos_t pos1;
    char hostname[MAXHOSTNAMELENGTH];
    char ipv4[MAXIPV4LENGTH];
    uint8_t mac[6];
    int fd;
    struct ifreq ifr;
    memset(hostname, '\0', sizeof ( hostname));
    memset(ipv4, '\0', sizeof (ipv4));
    memset(mac, '\0', sizeof (mac));

    pos1.Position.Longitude = 0;
    pos1.Position.Latitude = 0;
    pos1.Position.Altitude = 0;
    pos1.Position.Status = 1;

    for (i = 0; i < 6; i++) pos1.MAC[i] = 0;

    setCurrentStatusConfiguration(UAV_SCENARIO_NONE, UAV_SEARCHING_UNINITIALIZED, UAV_POSITIONING_UNINITIALIZED);


    setCurrentStatusState(UAV_STATE_UNCONFIGURED);
   
    setCurrentStatusClient1(pos1);
    setCurrentStatusClient2(pos1);

    gethostname(hostname, MAXHOSTNAMELENGTH);
    setCurrentStatusHostname(hostname);

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, bridgeinterface, IFNAMSIZ-1);
    ioctl(fd, SIOCGIFADDR, &ifr);
    
    sprintf( ipv4, "%s", inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr));
    printf("ipv4: %s\n", ipv4);
    setCurrentStatusIPv4(ipv4);
    
    ioctl(fd, SIOCGIFHWADDR, &ifr);
    close(fd);
    int j = 0;
    int k = 0;
    for (j=0; j<6; j++) {
        mac[j] = (unsigned char)ifr.ifr_hwaddr.sa_data[j];
    }
    
    setCurrentStatusMac(mac);

    result = SUCCESS;
    return result;
}

int handlingCommunication(int sock) {
    int n;
    int waitForSocketInput = TRUE;
    uint8_t command;
    char data[200];
    char inbuffer[MAXLINELENGTH];
    memset(inbuffer, '\0', MAXLINELENGTH);
    memset(data, '\0', sizeof ( data));
    int communicationHandlingFinished = FALSE;

    n = receiveTcpPacket(sock, inbuffer, MAXLINELENGTH);
    if (n == 0) {
        printf("caller closed the connection on socket %d\n", sock);
    } else if (n < 0) {
        printf("ERROR reading from socket\n");
    } else {
        printf("received %d bytes from socket: %s\n", n, inbuffer);
        printf("strlen(inbuffer): %d\n", strlen(inbuffer));
        if (checkCRC(inbuffer, strlen(inbuffer) - 1) == FAILURE) {
            printf("wrong crc\n");
            outSocketResponse(sock, UAV_RESPONSE_FAILURE);
            return FALSE;
        }
        inSocketCommand(inbuffer, strlen(inbuffer), &command, data);

        switch (command) {
            case 'w':
                //            inCompassData ( data );
                break;
            case 'W':
                //            inNbrOfWaypoints(data);
                break;
            case 'X':
                //            inWaypoint(data);
                break;
            case 'x':
                outSerialRequestWaypoint((int) * data, serialfd);
                break;
            case 'P':
                communicationHandlingFinished = inPositionSubmission(data, serialfd, &waitForSocketInput, sock);
                break;
            case 'D':
                communicationHandlingFinished = inFlightDirection(data, serialfd, sock);
                break;
            case 'N':
                communicationHandlingFinished = inNotificationSubscription(data, sock);
                break;
            case 'C':
                communicationHandlingFinished = inStartConfiguration(data, sock);
                break;
            case 'A':
                communicationHandlingFinished = inAbort(data, sock, serialfd);
                break;
            default:
                communicationHandlingFinished = UAV_RESPONSE_UNKNOWN;
        }
    }
    printf("handling communication finished\n");
    return communicationHandlingFinished;
}

int inAbort(char *data, int sockfd, int serialfd) {
    int abortdata = (int) data[0];
    int success = SUCCESS;
    extern int aborted;

    printf("aborting...\n");
    GPS_Pos_t position;
    memset(&position, '\0', sizeof (GPS_Pos_t));
    position = getHomePosition();

    outSerialSendWaypoint(position, FALSE, 1, serialfd);
    outSerialSendWaypoint(position, TRUE, 1, serialfd);
    outSerialSendWaypoint(position, TRUE, 2, serialfd);
    setCurrentStatusState(UAV_STATE_ABORTED);
    aborted = TRUE;
    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
    if (success == SUCCESS) {
        return TRUE;
    } else {
        return FALSE;
    }
}

GPS_Pos_t getHomePosition() {
    GPS_Pos_t homePosition;
    NaviData_t navidata;
    memset(&homePosition, '\0', sizeof (GPS_Pos_t));
    memset(&navidata, '\0', sizeof (NaviData_t));

    pthread_mutex_lock(&currentNaviDataMutex);
    navidata = currentNaviData;
    pthread_mutex_unlock(&currentNaviDataMutex);

    homePosition = navidata.HomePosition;

    return homePosition;
}

int inStartConfiguration(char *data, int sockfd) {
  
    uint8_t success = SUCCESS;
    uint8_t currentState = 0;
    
    StartConfig_t startConfig;
    memset(&startConfig, '\0', sizeof (StartConfig_t));
    
    memcpy((StartConfig_t*) & startConfig, (uint8_t *) data, sizeof (StartConfig_t));

    currentState = getCurrentStatusState();
    if (currentState == UAV_STATE_UNCONFIGURED || currentState == UAV_STATE_ABORTED) { //set configuration only if uav is unconfigured or aborted
	if(currentState == UAV_STATE_ABORTED) {
	  initStatus();
	}
        printf("state unconfigured\n");
        setCurrentStatusConfiguration(startConfig.Scenario, startConfig.Searching, startConfig.Positioning);
	if(startConfig.Scenario == UAV_SCENARIO_AIRBORNE || startConfig.Scenario == UAV_SCENARIO_MULTIHOP) {
	    setCurrentAllowedClients(startConfig.AllowedClients);
	}
        if (startConfig.Scenario == UAV_SCENARIO_AIRBORNE) {
            switch (startConfig.Searching) {
                case UAV_SEARCHING_AUTONOMOUS:
                    setCurrentStatusState(UAV_STATE_AIRBORNE_AWAITING_POS1);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
                    break;
                case UAV_SEARCHING_MANUAL:
                    setCurrentStatusState(UAV_STATE_AIRBORNE_AWAITING_FLIGHT_DIRECTION);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_FLIGHT_DIRECTION_REQUEST);
                    break;
            }
        } else if(startConfig.Scenario == UAV_SCENARIO_MULTIHOP) {
	    switch (startConfig.Searching) {
                case UAV_SEARCHING_AUTONOMOUS:
                    setCurrentStatusState(UAV_STATE_MULTIHOP_1_AWAITING_POS1);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
                    break;
                case UAV_SEARCHING_MANUAL:
                    setCurrentStatusState(UAV_STATE_MULTIHOP_1_AWAITING_FLIGHT_DIRECTION);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_FLIGHT_DIRECTION_REQUEST);
                    break;
	    }
	} else {
            success = outSocketResponse(sockfd, UAV_RESPONSE_UNKNOWN);
        }
    } else {
        success = outSocketResponse(sockfd, UAV_RESPONSE_FAILURE);
    }
    if (success == SUCCESS) {
        return TRUE;
    } else {
        return FALSE;
    }
}

int inNotificationSubscription(char *data, int sockfd) {

    int pos = -1;
    int success = FAILURE;

    NotificationSubscription_t notificationSubscription;
    memset(&notificationSubscription, 0, sizeof(NotificationSubscription_t));

    memcpy((NotificationSubscription_t*) & notificationSubscription, (uint8_t *) data, sizeof ( NotificationSubscription_t));

    notificationSubscription.Addr.ss_family = ntohs(notificationSubscription.Addr.ss_family);
    pos = containsSubscription(notificationSubscription);
    printf("pos: %d\n", pos);
    printf("notificationcounter: %d\n", notificationSubscriberCounter);

    if (notificationSubscription.Subscription == UAV_SUBSCRIPTION_SUBSCRIBE) {
        printf("notification subscription received\n");

        if (pos < 0) {
            if (notificationSubscriberCounter >= MAX_NOTIFICATION_SUBSCRIBER) {
                printf("counter > maxnotificationsubscriber\n");
                success = outSocketResponse(sockfd, UAV_RESPONSE_FAILURE);
            } else {
                notificationSubscriber[notificationSubscriberCounter] = notificationSubscription.Addr;
                notificationSubscriberCounter++;
                if (notificationSubscription.Addr.ss_family == AF_INET) {
                    printf("added ipv4 subscription\n");
                    outputIp(&notificationSubscription.Addr);
                } else if (notificationSubscription.Addr.ss_family == AF_INET6) {
                    printf("added ipv6 subscription\n");
                    outputIp(&notificationSubscription.Addr);
                } else {
                    printf("inNotificationSubscription: unknown address family: %d\n", notificationSubscription.Addr.ss_family);
                }
                if (simulation == 1) { //start simulation only if script was started using the simulation parameter
                    if (startedSimulation == 0) {
                        initSimulationData();
                    }
                    startedSimulation = 1;
                }
                success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
            }
        } else {
            success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
        }
    } else if (notificationSubscription.Subscription == UAV_SUBSCRIPTION_UNSUBSCRIBE) {
        printf("notification unsubscription received\n");

        if (pos >= 0) {
            notificationSubscriberCounter--;
            for (; pos < notificationSubscriberCounter; pos++) {
                notificationSubscriber[pos] = notificationSubscriber[pos + 1];
            }
	    memset(notificationSubscriber + notificationSubscriberCounter, 0, sizeof (struct sockaddr));
        }
        success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
    } else {
        success = outSocketResponse(sockfd, UAV_RESPONSE_FAILURE);
    }

    if (success == SUCCESS) {
        return TRUE;
    } else {
        return FALSE;
    }
}

int containsSubscription(NotificationSubscription_t notificationSubscription) {
    int pos = -1;
    int i = 0;

    char ipstrNewAddr[INET6_ADDRSTRLEN];
    char ipstrTmpAddr[INET6_ADDRSTRLEN];

    struct sockaddr_storage newAddr = notificationSubscription.Addr;
    struct sockaddr_storage tmpAddr;
    for (i = 0; i < notificationSubscriberCounter; i++) {
        tmpAddr = notificationSubscriber[i];
        if (tmpAddr.ss_family == newAddr.ss_family) { //same family?
	    if (newAddr.ss_family == AF_INET) {
                struct sockaddr_in *ipv4New = (struct sockaddr_in *) & newAddr;
                struct sockaddr_in *ipv4Tmp = (struct sockaddr_in *) & tmpAddr;
                inet_ntop(newAddr.ss_family, ipv4New, ipstrNewAddr, sizeof ipstrNewAddr);
                inet_ntop(newAddr.ss_family, ipv4Tmp, ipstrTmpAddr, sizeof ipstrTmpAddr);
                if (strcmp(ipstrNewAddr, ipstrTmpAddr) == 0) {
                    printf("found ipv4\n");
                    pos = i;
                    break;
                }
            } else if (newAddr.ss_family == AF_INET6) {
                struct sockaddr_in6 *ipv6New = (struct sockaddr_in6 *) & newAddr;
                struct sockaddr_in6 *ipv6Tmp = (struct sockaddr_in6 *) & tmpAddr;
                inet_ntop(newAddr.ss_family, ipv6New, ipstrNewAddr, sizeof ipstrNewAddr);
                inet_ntop(newAddr.ss_family, ipv6Tmp, ipstrTmpAddr, sizeof ipstrTmpAddr);
                if (strcmp(ipstrNewAddr, ipstrTmpAddr) == 0) {
                    printf("found ipv6\n");
                    pos = i;
                    break;
                }
            } else {
	      
	    }
        }
    }
    return pos;
}

int outSerialDebugInterval(int fd, uint8_t interval) {
    int address = UAV_ADDRESS_ALL;
    unsigned char command = 'd';
    char output[MAXDATALENGTH];
    memset(output, '\0', sizeof ( output));
    int n = 0;
    char data[1] = {interval};
    generateInternalCommand(address, command, data, sizeof (data), output);
    n = write(fd, output, strlen(output));
    return n;
}

int outSerialOSDInterval(int fd, uint8_t interval) {
    int address = UAV_ADDRESS_NC;
    unsigned char command = 'o';
    char output[MAXDATALENGTH];
    memset(output, '\0', sizeof ( output));
    int n = 0;
    generateInternalCommand(address, command, &interval, sizeof (uint8_t), output);
    n = write(fd, output, strlen(output));
    return n;
}

int outSerialLinkTest(int fd, uint16_t pattern) {
    int address = UAV_ADDRESS_NC;
    unsigned char command = 'z';
    char output[MAXDATALENGTH];
    memset(output, '\0', sizeof ( output));
    int n = 0;
    generateInternalCommand(address, command, &pattern, sizeof (uint16_t), output);
    n = write(fd, output, strlen(output));
    return n;
}

void outSerialRequestWaypoint(unsigned char index, int fd) {
    int address = UAV_ADDRESS_NC;
    unsigned char command = 'x';
    char output[MAXDATALENGTH];
    memset(output, '\0', sizeof ( output));

    char data[1] = {index - 1};

    generateInternalCommand(address, command, data, sizeof ( data), output);
    write(fd, output, strlen(output));
}

void outSerialSelectNC(int fd) // n
{
    char testdata[5] = {0x1B, 0x1B, 0x55, 0xAA, 0x00};

    printf("redirecting UART to NC...\r\n");
    write(fd, testdata, 5);
}

int inPositionSubmission(char *data, int fd, int* waitForSocketInput, int sockfd) {
    Submitted_Pos_t position;
    int currentState;
    int currentSearching;
    int currentScenario;
    int success = FAILURE;
    double r = 0.010; //Spiral radius: 4m
    double start_deg = 0; //Spiral: Start to the north
    int max_turns = 3;
    double incr_deg = 90; //Spiral: increased degrees
    int nbrWaypoints = (360 / incr_deg * max_turns) + 1;
    
    printf("recieved position submission\n");

    memcpy((Submitted_Pos_t*) & position, (uint8_t *) data, sizeof ( Submitted_Pos_t));
    
    currentScenario = getCurrentStatusScenario();
    if(currentScenario == UAV_SCENARIO_AIRBORNE || currentScenario == UAV_SCENARIO_MULTIHOP) {
	if(isClientAllowed(position.MAC) == FALSE) {
	    printf("client not allowed: ");
	    printf("%.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", position.MAC[0], position.MAC[1], position.MAC[2], position.MAC[3], position.MAC[4], position.MAC[5]);
	    printf("allowed 1: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", currentAllowedClients.MAC1[0], currentAllowedClients.MAC1[1], currentAllowedClients.MAC1[2], currentAllowedClients.MAC1[3], currentAllowedClients.MAC1[4], currentAllowedClients.MAC1[5]);
	    printf("allowed 2: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", currentAllowedClients.MAC2[0], currentAllowedClients.MAC2[1], currentAllowedClients.MAC2[2], currentAllowedClients.MAC2[3], currentAllowedClients.MAC2[4], currentAllowedClients.MAC2[5]);
	    success = outSocketResponse(sockfd, UAV_RESPONSE_FAILURE);
	    if (success == SUCCESS) {
		return TRUE;
	    } else {
		return FALSE;
	    }
	}
    }
    
    position.Position.Longitude = endian_swap_32(position.Position.Longitude);
    position.Position.Latitude = endian_swap_32(position.Position.Latitude);
    position.Position.Altitude = endian_swap_32(position.Position.Altitude);
    
    printf("submittedPositionsCounter: %d\n", submittedPositionsCounter);
    printf("received position: lon: %ld, lat: %ld, altitude: %ld, status: %d\n", position.Position.Longitude, position.Position.Latitude, position.Position.Altitude, position.Position.Status);
    
    outputCurrentScenario();

    currentState = getCurrentStatusState();
    currentSearching = getCurrentStatusSearching();
    printf("currentState: %d\n", currentState);
    printf("currentSearching: %d\n", currentSearching);
    switch (currentState) {
        case UAV_STATE_AIRBORNE_AWAITING_POS1:
            printf("awaiting pos1\n");
            submittedPositions[0] = position;
            setCurrentStatusClient1(position);
            submittedPositionsCounter = 1;
            switch (currentSearching) {
                case UAV_SEARCHING_MANUAL:
                    printf("searching manual\n");
		    setCurrentStatusState(UAV_STATE_AIRBORNE_AWAITING_POS2);
                    outSerialSendWaypoint(flightDirectionPosition.Position, FALSE, 1, serialfd); //delete waypoint list with invalid waypoint
                    outSerialSendWaypoint(flightDirectionPosition.Position, TRUE, 1, serialfd); //send valid waypoint
                    outSerialSendWaypoint(flightDirectionPosition.Position, TRUE, 2, serialfd); //send valid waypoint
                    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
                    break;
                case UAV_SEARCHING_AUTONOMOUS:
                    setCurrentStatusState(UAV_STATE_AIRBORNE_AWAITING_POS2);
                    outSerialSendWaypoint(position.Position, FALSE, 1, serialfd); //delete list
                    GPS_Pos_t spiralWaypoints[nbrWaypoints];
                    GPS_Pos_t tmpWaypoint;
                    calcSpiralWaypoints(spiralWaypoints, position.Position, r, incr_deg, start_deg, max_turns);
                    int i = 0;
                    for (i = 0; i < nbrWaypoints; i++) {
                        tmpWaypoint = spiralWaypoints[i];
                        printf("waypoint %d: lat: %f, lon: %f\n", i, ((double) tmpWaypoint.Latitude / 10000000), ((double) tmpWaypoint.Longitude / 10000000));
			outSerialSendWaypoint(tmpWaypoint, TRUE, i+1, serialfd);
                    }
		    outSerialSendWaypoint(tmpWaypoint, TRUE, i+1, serialfd);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
                    break;
            }
	    break;
	case UAV_STATE_MULTIHOP_1_AWAITING_POS1:
	    printf("awaiting pos1\n");
            submittedPositions[0] = position;
            setCurrentStatusClient1(position);
            submittedPositionsCounter = 1;
            switch (currentSearching) {
                case UAV_SEARCHING_MANUAL:
                    printf("searching manual\n");
                    outSerialSendWaypoint(flightDirectionPosition.Position, FALSE, 1, serialfd); //delete waypoint list with invalid waypoint
                    outSerialSendWaypoint(flightDirectionPosition.Position, TRUE, 1, serialfd); //send valid waypoint
                    outSerialSendWaypoint(flightDirectionPosition.Position, TRUE, 2, serialfd); //send valid waypoint
                    setCurrentStatusState(UAV_STATE_MULTIHOP_1_AWAITING_POS2);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
                    break;
                case UAV_SEARCHING_AUTONOMOUS:
                    outSerialSendWaypoint(position.Position, FALSE, 1, serialfd); //delete list
                    GPS_Pos_t spiralWaypoints[nbrWaypoints];
                    GPS_Pos_t tmpWaypoint;
                    calcSpiralWaypoints(spiralWaypoints, position.Position, r, incr_deg, start_deg, max_turns);
                    int i = 0;
                    for (i = 0; i < nbrWaypoints; i++) {
                        tmpWaypoint = spiralWaypoints[i];
                        printf("waypoint %d: lat: %f, lon: %f\n", i, ((double) tmpWaypoint.Latitude / 10000000), ((double) tmpWaypoint.Longitude / 10000000));
                        outSerialSendWaypoint(tmpWaypoint, TRUE, i + 1, serialfd);
                    }
		    outSerialSendWaypoint(tmpWaypoint, TRUE, i + 1, serialfd);
                    setCurrentStatusState(UAV_STATE_MULTIHOP_1_AWAITING_POS2);
                    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
                    break;
            }
	    break;
        case UAV_STATE_AIRBORNE_AWAITING_POS2:
            if (newMAC(position.MAC, submittedPositions[0].MAC) == TRUE) {
                submittedPositions[1] = position;
                setCurrentStatusClient2(position);
                calculateCenterPosition();
                outSerialSendWaypoint(submittedPositions[2].Position, FALSE, 1, serialfd);
                outSerialSendWaypoint(submittedPositions[2].Position, TRUE, 1, serialfd);
                outSerialSendWaypoint(submittedPositions[2].Position, TRUE, 2, serialfd);
		setCurrentStatusState(UAV_STATE_AIRBORNE_FLYING_CENTER);
                success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
            }
            break;
	case UAV_STATE_MULTIHOP_1_AWAITING_POS2:
            if (newMAC(position.MAC, submittedPositions[0].MAC) == TRUE) {
                submittedPositions[1] = position;
                setCurrentStatusClient2(position);
                calculateCenterPosition();
                outSerialSendWaypoint(submittedPositions[2].Position, FALSE, 1, serialfd);
                outSerialSendWaypoint(submittedPositions[2].Position, TRUE, 1, serialfd);
                outSerialSendWaypoint(submittedPositions[2].Position, TRUE, 2, serialfd);
		setCurrentStatusState(UAV_STATE_MULTIHOP_1_FLYING_CENTER);
		success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);
            }
            break;
    }

    if (success == SUCCESS) {
        return TRUE;
    } else {
        return FALSE;
    }
}

int getCurrentStatusScenario() {
    int scenario;
    pthread_mutex_lock(&currentStatusMutex);
    scenario = currentStatus.Scenario;
    pthread_mutex_unlock(&currentStatusMutex);
    return scenario;
}

int getCurrentStatusState() {
    int state;
    pthread_mutex_lock(&currentStatusMutex);
    state = currentStatus.State;
    pthread_mutex_unlock(&currentStatusMutex);
    return state;
}

int getCurrentStatusSearching() {
    int searching;
    pthread_mutex_lock(&currentStatusMutex);
    searching = currentStatus.Searching;
    pthread_mutex_unlock(&currentStatusMutex);
    return searching;
}

int getCurrentStatusPositioning() {
    int positioning;
    pthread_mutex_lock(&currentStatusMutex);
    positioning = currentStatus.Positioning;
    pthread_mutex_unlock(&currentStatusMutex);
    return positioning;
}

int inFlightDirection(char *data, int fd, int sockfd) {
    printf("flight direction received\n");
    int serialInput = 0;
    int success = FAILURE;

    Submitted_Pos_t position;
    memset(&position, '\0', sizeof ( Submitted_Pos_t));
    
    memcpy((Submitted_Pos_t*) & position, data, sizeof ( Submitted_Pos_t));
    
    position.Position.Longitude = endian_swap_32(position.Position.Longitude);
    position.Position.Latitude = endian_swap_32(position.Position.Latitude);
    position.Position.Altitude = endian_swap_32(position.Position.Altitude);
    
    memset(&flightDirectionPosition, '\0', sizeof ( Submitted_Pos_t));
    flightDirectionPosition = position;
    
    if(getCurrentStatusState() == UAV_STATE_AIRBORNE_AWAITING_FLIGHT_DIRECTION) {
	setCurrentStatusState(UAV_STATE_AIRBORNE_AWAITING_POS1);
    } else if(getCurrentStatusState() == UAV_STATE_MULTIHOP_1_AWAITING_FLIGHT_DIRECTION) {
	setCurrentStatusState(UAV_STATE_MULTIHOP_1_AWAITING_POS1);
    }

    outSerialSendWaypoint(position.Position, TRUE, 1, fd);
    outSerialSendWaypoint(position.Position, TRUE, 2, fd);

    success = outSocketResponse(sockfd, UAV_RESPONSE_SUCCESS);

    if (success == SUCCESS) {
        return TRUE;
    } else {
        return FALSE;
    }
}

void outSerialSelectFC(int fd) // f
{
    char output[200];
    memset(output, '\0', sizeof ( output));
    unsigned char command = 'u';

    char data[1] = {0};

    generateInternalCommand(UAV_ADDRESS_NC, command, data, sizeof ( data), output);
    printf("redirecting UART to FC... ");
    puts(output);
    write(fd, output, strlen(output));
}

void outSerialSendWaypoint(GPS_Pos_t position, int correct, uint8_t wpindex, int fd) {

    listeningForSerialInput = TRUE;
    outSerialSelectNC(fd);
    char output[MAXLINELENGTH];
    unsigned char command = 'w';
    char data[MAXDATALENGTH];
    memset(output, '\0', sizeof ( output));
    memset(data, '\0', sizeof ( data));

    int16_t heading = 0;
    uint8_t toleranceRadius = 4;
    uint8_t holdTime = 2;
    uint8_t eventFlag = 0;
    uint8_t index = wpindex;
    if(correct == FALSE) {
      index = 0;
    }

    Waypoint_t target;
    target.Position = position;
    
    printf("target position: lon: %ld, lat: %ld, altitude: %ld, status: %d\n", target.Position.Longitude, target.Position.Latitude, target.Position.Altitude, target.Position.Status);
    target.Position.Longitude = endian_swap_32(target.Position.Longitude);
    target.Position.Latitude = endian_swap_32(target.Position.Latitude);
    target.Position.Altitude = endian_swap_32(target.Position.Altitude);
    target.Heading = endian_swap_16(target.Heading);
    
    if (correct == TRUE) {
        target.Position.Status = 1;
    } else if (correct == FALSE) {
        target.Position.Status = 0;
    }
    target.Heading = heading;
    target.ToleranceRadius = toleranceRadius;
    target.HoldTime = holdTime;
    target.Event_Flag = eventFlag;
    target.Index = index;
    target.Reserve[0] = 0;
    target.Reserve[1] = 0;
    target.Reserve[2] = 0;
    target.Reserve[3] = 0;
    target.Reserve[4] = 0;
    target.Reserve[5] = 0;
    target.Reserve[6] = 0;
    target.Reserve[7] = 0;
    target.Reserve[8] = 0;
    target.Reserve[9] = 0;
    target.Reserve[10] = 0;
    
    memcpy((uint8_t *) & data, (uint8_t *) & target, sizeof ( Waypoint_t));
    generateInternalCommand(UAV_ADDRESS_NC, command, data, sizeof ( Waypoint_t), output);
    write(fd, output, strlen(output));
}

void outSerialSendTarget(GPS_Pos_t position, int correct, uint8_t wpindex, int fd) {

    listeningForSerialInput = TRUE;
    outSerialSelectNC(fd);

    char output[MAXLINELENGTH];
    unsigned char command = 's';
    char data[MAXDATALENGTH];
    memset(output, '\0', sizeof ( output));
    memset(data, '\0', sizeof ( data));

    int16_t heading = 0;
    uint8_t toleranceRadius = 2;
    uint8_t holdTime = 10;
    uint8_t eventFlag = 0;
    uint8_t index = wpindex;
    
    if(correct == FALSE) {
      index = 0;
    }

    Waypoint_t target;
    target.Position = position;
    
    printf("target: lon: %ld, lat: %ld, altitude: %ld, status: %d\n", target.Position.Longitude, target.Position.Latitude, target.Position.Altitude, target.Position.Status);
	   
    target.Position.Longitude = endian_swap_32(target.Position.Longitude);
    target.Position.Latitude = endian_swap_32(target.Position.Latitude);
    target.Position.Altitude = endian_swap_32(target.Position.Altitude);
    target.Heading = endian_swap_16(target.Heading);
    
    if (correct == TRUE) {
        target.Position.Status = 1;
    } else if (correct == FALSE) {
        target.Position.Status = 0;
    }
    target.Heading = heading;
    target.ToleranceRadius = toleranceRadius;
    target.HoldTime = holdTime;
    target.Event_Flag = eventFlag;
    target.Index = index;
    target.Reserve[0] = 0;
    target.Reserve[1] = 0;
    target.Reserve[2] = 0;
    target.Reserve[3] = 0;
    target.Reserve[4] = 0;
    target.Reserve[5] = 0;
    target.Reserve[6] = 0;
    target.Reserve[7] = 0;
    target.Reserve[8] = 0;
    target.Reserve[9] = 0;
    target.Reserve[10] = 0;    
    
    memcpy((uint8_t *) & data, (uint8_t *) & target, sizeof ( Waypoint_t));

    generateInternalCommand(UAV_ADDRESS_NC, command, data, sizeof ( Waypoint_t), output);

    write(fd, output, strlen(output));
}

int outSocketResponse(int sockfd, int response) {
    char output[MAXLINELENGTH];
    unsigned char command = 'r';
    int n = 0;

    memset(output, '\0', sizeof (output));

    char data[1] = {response};
    generateExternalCommand(command, data, sizeof (data), output);
    printf("sending response: %d: %s\n", response, output);
    n = sendTcpPacketOnExistingSocket(sockfd, output, strlen(output), FALSE, 0);
    if (n > 0) {
        return SUCCESS;
    } else {
        return FAILURE;
    }
}

void calculateCenterPosition() {
    int32_t altitude = 10000; //10m
    uint8_t status = 1;

    submittedPositions[2].Position.Longitude = (submittedPositions[0].Position.Longitude + submittedPositions[1].Position.Longitude) / 2;
    submittedPositions[2].Position.Latitude = (submittedPositions[0].Position.Latitude + submittedPositions[1].Position.Latitude) / 2;
    submittedPositions[2].Position.Altitude = altitude;
    submittedPositions[2].Position.Status = status;
}

void setCurrentStatusConfiguration(int scenario, int searching, int positioning) {
    printf("setCurrentStatusConfiguration: scenario: %d searching: %d positioning: %d\n", scenario, searching, positioning);
    pthread_mutex_lock(&currentStatusMutex);
    currentStatus.Scenario = scenario;
    currentStatus.Searching = searching;
    currentStatus.Positioning = positioning;
    pthread_mutex_unlock(&currentStatusMutex);
}

void setCurrentAllowedClients(AllowedClients_t allowedClients) {
    pthread_mutex_lock(&currentAllowedClientsMutex);
    currentAllowedClients = allowedClients;
    pthread_mutex_unlock(&currentAllowedClientsMutex);
}

AllowedClients_t getCurrentAllowedClients() {
    AllowedClients_t allowedClients;
    memset(&allowedClients, '\0', sizeof(AllowedClients_t));
    pthread_mutex_lock(&currentAllowedClientsMutex);
    allowedClients = currentAllowedClients;
    pthread_mutex_unlock(&currentAllowedClientsMutex);
    return allowedClients;    
}

int isClientAllowed(uint8_t client[]) {
    int allowed = TRUE;
    int i = 0;
    AllowedClients_t allowedClients;
    memset(&allowedClients, '\0', sizeof(AllowedClients_t));
    allowedClients = getCurrentAllowedClients();
    for(i=0; i<6; i++) {
	if(allowedClients.MAC1[i] != client[i]) {
	  allowed = FALSE;
	  break;
	}
    }
    if(allowed == TRUE) {
	return allowed;
    } else {
      allowed = TRUE;
    }
    for(i=0; i<6; i++) {
	if(allowedClients.MAC2[i] != client[i]) {
	  allowed = FALSE;
	  break;
	} 
    }
    return allowed;
}

void setCurrentNaviData(NaviData_t navidata) {
    pthread_mutex_lock(&currentNaviDataMutex);
    currentNaviData = navidata;
    pthread_mutex_unlock(&currentNaviDataMutex);
}

void setCurrentStatusPosition(NaviData_t navidata) {
    pthread_mutex_lock(&currentStatusMutex);
    currentStatus.Position = navidata.CurrentPosition;
    pthread_mutex_unlock(&currentStatusMutex);
}

void setCurrentStatusState(int state) {
    printf("\n##  setting state %d\n\n", state);
    pthread_mutex_lock(&currentStatusMutex);
    currentStatus.State = state;
    pthread_mutex_unlock(&currentStatusMutex);

}

void setCurrentStatusClient1(Submitted_Pos_t position) {
    pthread_mutex_lock(&currentStatusMutex);
    currentStatus.Client1 = position;
    pthread_mutex_unlock(&currentStatusMutex);
}

void setCurrentStatusClient2(Submitted_Pos_t position) {
    pthread_mutex_lock(&currentStatusMutex);
    currentStatus.Client2 = position;
    pthread_mutex_unlock(&currentStatusMutex);
}

Submitted_Pos_t getCurrentStatusClient1() {
    Submitted_Pos_t position;
    memset(&position, '\0', sizeof(Submitted_Pos_t));
    pthread_mutex_lock(&currentStatusMutex);
    position = currentStatus.Client1;
    pthread_mutex_unlock(&currentStatusMutex);
    return position;
}

Submitted_Pos_t getCurrentStatusClient2() {
    Submitted_Pos_t position;
    memset(&position, '\0', sizeof(Submitted_Pos_t));
    pthread_mutex_lock(&currentStatusMutex);
    position = currentStatus.Client2;
    pthread_mutex_unlock(&currentStatusMutex);
    return position;
}

void setCurrentStatusHostname(char * hostname) {
    pthread_mutex_lock(&currentStatusMutex);
    strcpy(currentStatus.Hostname, hostname);
    pthread_mutex_unlock(&currentStatusMutex);
}

void setCurrentStatusIPv4(char * ipv4) {
    pthread_mutex_lock(&currentStatusMutex);
    strcpy(currentStatus.IPv4, ipv4);
    pthread_mutex_unlock(&currentStatusMutex);
}

void setCurrentStatusMac(char * mac) {
    int i = 0;  
    pthread_mutex_lock(&currentStatusMutex);
    for(i=0; i<6; i++) {
      currentStatus.MAC[i] = mac[i];
    }
    pthread_mutex_unlock(&currentStatusMutex);
}

void outputCurrentScenario() {
    pthread_mutex_lock(&currentStatusMutex);
    pthread_mutex_unlock(&currentStatusMutex);
}

void getCurrentStatusMac(char * mac) {
    pthread_mutex_lock(&currentStatusMutex);
    sprintf(mac, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", currentStatus.MAC[0], currentStatus.MAC[1], currentStatus.MAC[2], currentStatus.MAC[3], currentStatus.MAC[4], currentStatus.MAC[5]);
    pthread_mutex_unlock(&currentStatusMutex);
}

void getCurrentClient1MacString(char * mac) {
    pthread_mutex_lock(&currentStatusMutex);
    sprintf(mac, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x",currentStatus.Client1.MAC[0], currentStatus.Client1.MAC[1], currentStatus.Client1.MAC[2], currentStatus.Client1.MAC[3], currentStatus.Client1.MAC[4], currentStatus.Client1.MAC[5]);
    pthread_mutex_unlock(&currentStatusMutex);
}

void getCurrentClient2MacString(char * mac) {
    pthread_mutex_lock(&currentStatusMutex);
    sprintf(mac, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x",currentStatus.Client2.MAC[0], currentStatus.Client2.MAC[1], currentStatus.Client2.MAC[2], currentStatus.Client2.MAC[3], currentStatus.Client2.MAC[4], currentStatus.Client2.MAC[5]);
    pthread_mutex_unlock(&currentStatusMutex);
}

void outputCurrentClient1() {
    pthread_mutex_lock(&currentStatusMutex);
    printf("client1: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", currentStatus.Client1.MAC[0], currentStatus.Client1.MAC[1], currentStatus.Client1.MAC[2], currentStatus.Client1.MAC[3], currentStatus.Client1.MAC[4], currentStatus.Client1.MAC[5]);
    printf("client1: lon: %d, lat: %d, altitude: %d, status: %d\n", currentStatus.Client1.Position.Longitude, currentStatus.Client1.Position.Latitude, currentStatus.Client1.Position.Altitude, currentStatus.Client1.Position.Status);
    pthread_mutex_unlock(&currentStatusMutex);
}
void outputCurrentClient2() {
    pthread_mutex_lock(&currentStatusMutex);
    printf("client2: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", currentStatus.Client2.MAC[0], currentStatus.Client2.MAC[1], currentStatus.Client2.MAC[2], currentStatus.Client2.MAC[3], currentStatus.Client2.MAC[4], currentStatus.Client2.MAC[5]);
    printf("client2: lon: %d, lat: %d, altitude: %d, status: %d\n", currentStatus.Client2.Position.Longitude, currentStatus.Client2.Position.Latitude, currentStatus.Client2.Position.Altitude, currentStatus.Client2.Position.Status);
    pthread_mutex_unlock(&currentStatusMutex);
}

int checkInputs(int argc, char** argv) {
    int success = SUCCESS;
    int c;

    serialdevice = DEFAULTSERIALDEVICE;

    while ((c = getopt(argc, argv, "t:s")) != -1)
        switch (c) {
            case 't':
                serialdevice = optarg;
                break;
            case 's':
                simulation = 1;
                break;
            default:
                usage(argv[0]);
                success = FAILURE;
        }

    return success;
}

void usage(char * self) {
    printf("usage: %s [-t serialdevice] [-s]\n", self);
    printf("  -t serial device   Use this serial device instead of the default serial device %s\n", DEFAULTSERIALDEVICE);
    printf("  -s                 Use simulated data instead of real data\n");
    printf("example: %s -t /dev/ttyS1 -s (using serial device ttyS1, simulation)\n", self);
    exit(EXIT_FAILURE);
}

void getConfigVariables() {
  
    char line[300];
    const char delimiters[] = "\"=";
    char *token;
    
    const char filename[] = "/etc/conf.d/network.conf";
    char firstword[100];
    char secondword[100];
    
    memset(line, '\0', sizeof(line));
    memset(meshinterface, '\0', sizeof(meshinterface));
    memset(wlaninterface, '\0', sizeof(wlaninterface));
    memset(bridgeinterface, '\0', sizeof(bridgeinterface));
    
    FILE *file = fopen(filename, "r");
    if (file == NULL) {
        printf("cannot open file\n");
	fclose(file) ;
        exit(1);
    }
    
    while (fgets(line, sizeof (line), file) != NULL) {
	memset(firstword, '\0', sizeof(firstword));
	memset(secondword, '\0', sizeof(secondword));
    
	token = strtok(line, delimiters);
	if(token != NULL) {
	  strcpy(firstword, token);
	}
	token = strtok(NULL, delimiters);
	if(token != NULL) {
	  strcpy(secondword, token);
	}
	if(strcmp(firstword, "UAVNET_MESH_DEVICE") == 0) {
	  strcpy(meshinterface, secondword);
	}
	if(strcmp(firstword, "UAVNET_WLAN_DEVICE") == 0) {
	  strcpy(wlaninterface, secondword);
	}
	if(strcmp(firstword, "UAVNET_BRIDGE_DEVICE") == 0) {
	  strcpy(bridgeinterface, secondword);
	}
    }
    fclose(file);
}
