#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <net/if.h>

#include "types.h"
#include "constants.h"
#include "notifier.h"
#include "simulator.h"

#define NOTIFICATIONINTERVAL 3

void * notify(void *iface) {

    extern int notificationSubscriberCounter;
    extern struct sockaddr_storage notificationSubscriber[MAX_NOTIFICATION_SUBSCRIBER];

    unsigned char command = 'n';
    Notification_t notification;
    int sockfd_udp, ifn = 0;
    int debug = 0;
    int i = 0;
    char payload[MAXBUFLEN];
    char output[MAXLINELENGTH];

    char charPort[5];
    memset(charPort, 0, sizeof charPort);
    sprintf(charPort, "%d", UAV_PORT_NOTIFICATION);

    while (ifn == 0) {
        ifn = if_nametoindex(iface);

        if (ifn == 0) {
            if (debug)
                perror("if_nametoindex");
            printf("Interface %s not available (yet) - sleeping\n",
                    (char*) iface);
            sleep(1);
        }
    }

    while (1) {
	
	for (i = 0; i < notificationSubscriberCounter && i < MAX_NOTIFICATION_SUBSCRIBER; i++) {
            sockfd_udp = generateControllerNotificationSocket(ifn, notificationSubscriber[i], charPort);
	    getNotificationValues(&notification);
        
            if (sockfd_udp < 0) {
                printf("error: open_socket");
                return;
            }

            memset(payload, '\0', sizeof (payload));
            memset(output, '\0', sizeof ( output));
            
	    endian_swap_notification(&notification);
	    
	    memcpy((uint8_t *) & payload, (uint8_t *) & notification, sizeof ( Notification_t));
            
	    generateExternalCommand(command, payload, sizeof (Notification_t), output);

            if (sendNotification(sockfd_udp, output) == -1) {
                printf("error: notifier.c sendto\n");
                return;
            }
            closeSocket(sockfd_udp);
        }
        sleep(NOTIFICATIONINTERVAL);
    }
}

void getNotificationValues(Notification_t * notification) {

    extern int startedSimulation;
    extern NaviData_t currentNaviData;
    extern pthread_mutex_t currentNaviDataMutex;

    NaviData_t navidata;

    if (startedSimulation == 1) {
        getSimulatedNotificationValues(notification);
    } else {
        
	pthread_mutex_lock(&currentNaviDataMutex);
        navidata = currentNaviData;
        pthread_mutex_unlock(&currentNaviDataMutex);


        extern Status_t currentStatus;
        extern pthread_mutex_t currentStatusMutex;

        notification->Position = navidata.CurrentPosition;

        notification->Height = navidata.Altimeter;
        notification->Battery = navidata.UBat; //in 0.1Volts
        notification->Heading = navidata.Heading; //in 1°
        notification->Speed = navidata.GroundSpeed; // in cm/s,  100cm/s -> 10m/s -> 36km/h
        pthread_mutex_lock(&currentStatusMutex);
        notification->Status = currentStatus;
        pthread_mutex_unlock(&currentStatusMutex);
	
    }
    //outputNotification(notification);
}

void outputNotification(Notification_t * notification) {

    printf("notificationlistener: packet contains:\n");
    printf("Position.Longitude: %i\n", notification->Position.Longitude);
    printf("Position.Latitude: %i\n", notification->Position.Latitude);
    printf("Position.Altitude: %d\n", notification->Position.Altitude);
    printf("Position.Status: %d\n", notification->Position.Status);
    printf("Height: %hd\n", notification->Height);
    printf("battery: %d\n", notification->Battery);
    printf("Heading: %d\n", notification->Heading);
    printf("Speed: %d\n", notification->Speed);
    printf("Status.Scenario: %u\n", notification->Status.Scenario);
    printf("Status.State: %u\n", notification->Status.State);
    printf("Status.Positioning: %u\n", notification->Status.Positioning);
    printf("Status.Searching: %u\n", notification->Status.Searching);
}
