#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "constants.h"
#include "types.h"
#include "socket.h"
#include "pinglistener.h"
#include "controller.h"

void * listen_for_pings(void *iface) {

    extern int pingReceived;
    extern Status_t currentStatus;
    extern pthread_mutex_t currentStatusMutex;
    
    int pingSocket;

    int numbytes;
    char buf[MAXBUFLEN];
    char sender[INET6_ADDRSTRLEN];
    Status_t status;
    uint8_t command;
    char data[MAXDATALENGTH];

    pingSocket = generateUdpPingListeningSocket(UAV_PORT_PING, (char*)iface);

    while (1) {

        memset(buf, 0, sizeof (buf));
        memset(sender, 0, sizeof (sender));
	
	printf("pinglistener: waiting to receive PING from uav on socket %d\n", pingSocket);

        numbytes = receiveUdpPacket(pingSocket, buf, sender, sizeof(sender));
	if (numbytes == -1) {
            printf("error: pinglistener.c recvfrom");
        } 
        if (checkCRC(buf, strlen(buf) - 1) == FAILURE) {
            printf("wrong crc\n");
        } else {
            inSocketCommand(buf, strlen(buf), &command, data);
            switch (command) {
                case 'p':
                    memcpy((Status_t*) & status, (uint8_t *) data, sizeof ( Status_t));
		    endian_swap_status(&status);
                    pingReceived = TRUE;
		    handlePing(status);
                    break;
                default:
                    printf("wrong command\n");
            }
        }
    }
    closeSocket(pingSocket);
}

void handlePing(Status_t status) {
  
    extern Status_t currentStatus;
    extern pthread_mutex_t currentStatusMutex;
    extern int serialfd;
    extern int aborted;
    GPS_Pos_t position;
	
    
    Status_t mystatus;    
    memset(&mystatus, '\0', sizeof(Status_t));
    
    pthread_mutex_lock(&currentStatusMutex);
    mystatus = currentStatus;
    pthread_mutex_unlock(&currentStatusMutex);
    
    if(strcmp(mystatus.Hostname, status.Hostname) == 0) {
      //my own ping, do nothing
    } else {
      outputPing(status);
      printf("handling ping from uav %s\n", status.Hostname);
      registerUav(status);
      updateClientLocations(status);
      if(status.State == UAV_STATE_ABORTED && mystatus.State != UAV_STATE_ABORTED && aborted == FALSE) {
	memset(&position, '\0', sizeof (GPS_Pos_t));
	position = getHomePosition();
	outSerialSendWaypoint(position, FALSE, 1, serialfd);
	outSerialSendWaypoint(position, TRUE, 1, serialfd);
	outSerialSendWaypoint(position, TRUE, 2, serialfd);
	setCurrentStatusState(UAV_STATE_ABORTED);
	aborted = TRUE;
	printf("aborted...\n");
      }
    }
}

void registerUav(Status_t status) {
    extern Status_t uavs[MAX_UAVS];
    extern int uavscounter;
    extern pthread_mutex_t uavsMutex;
    extern int stateFixCounter;

    Status_t tmpUav;
    int i = 0;
    int uavAlreadyRegistered = 0;
    uint8_t currentState = getCurrentStatusState();
    memset(&tmpUav, 0, sizeof(Status_t));
    int sameStateFound = 0;
    
    pthread_mutex_lock(&uavsMutex);
    for(i=0; i<uavscounter; i++) {
	if(i>=MAX_UAVS) {
	  break;
	}
	tmpUav = uavs[i];
	if(currentState == tmpUav.State && (currentState == UAV_STATE_MULTIHOP_2_WAITING || currentState == UAV_STATE_MULTIHOP_3_WAITING)) {
		sameStateFound = 1;
	    }
	    
	if(strcmp(tmpUav.Hostname, status.Hostname) == 0) {    
	    uavAlreadyRegistered = 1;
	    uavs[i] = status; //update uav
	}
    }
	    
    if(sameStateFound == 1 && stateFixCounter == -1) {
	srand (time (0));
	stateFixCounter = rand()%MAX_STATE_FIXER_PINGS + 1;
    } else if(sameStateFound == 0 && stateFixCounter >= 0) {
	stateFixCounter = -1;
    } else if(stateFixCounter > 0){
        stateFixCounter--;
    }
	    
    if(uavAlreadyRegistered==0) {
	printf("registering new uav...\n");
	if(uavscounter < MAX_UAVS) {
	  uavs[uavscounter++] = status;
	}
    }
    pthread_mutex_unlock(&uavsMutex);
	    
}

void updateClientLocations(Status_t status) {
    if(status.State == UAV_STATE_MULTIHOP_1_AWAITING_FLIGHT_DIRECTION ||
       status.State == UAV_STATE_MULTIHOP_1_AWAITING_POS2) {
	  setCurrentStatusClient1(status.Client1);
    }
    if(status.State == UAV_STATE_MULTIHOP_1_FLYING_CENTER ||
       status.State == UAV_STATE_MULTIHOP_1_REACHED_CENTER ||
       status.State == UAV_STATE_MULTIHOP_1_FLYING_FINAL_POSITION ||
       status.State == UAV_STATE_MULTIHOP_1_REACHED_FINAL_POSITION) {
	  setCurrentStatusClient2(status.Client2);
    }
}

void outputPing(Status_t status) {
    printf("scenario: %d\n", status.Scenario);
    printf("state: %d\n", status.State);
    printf("Position: lon: %ld, lat: %ld, altitude: %ld, status: %d\n", status.Position.Longitude, status.Position.Latitude, status.Position.Altitude, status.Position.Status);
    printf("client1: lon: %ld, lat: %ld, altitude: %ld, status: %d\n", status.Client1.Position.Longitude, status.Client1.Position.Latitude, status.Client1.Position.Altitude, status.Client1.Position.Status);
    printf("client2: lon: %ld, lat: %ld, altitude: %ld, status: %d\n", status.Client2.Position.Longitude, status.Client2.Position.Latitude, status.Client2.Position.Altitude, status.Client2.Position.Status);
    printf("hostname: %s\n", status.Hostname);
    printf("ipv4: %s\n", status.IPv4);
}
