#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include "types.h"
#include "constants.h"
#include "simulator.h"

#define MAXSIMULATIONSTEPS 100
#define DATAFILENAME "/sbin/SimulationData.csv"
#define MAXDATAFILELINELENGTH 1024

Notification_t notifications[MAXSIMULATIONSTEPS];
int simulationSteps = 0;

void initSimulationData() {
    Notification_t notification;
    GPS_Pos_t position, posCl1, posCl2;
    Status_t simStatus;
    extern Status_t currentStatus;
    extern pthread_mutex_t currentStatusMutex;
    Submitted_Pos_t client1, client2;

    char hostname[MAXHOSTNAMELENGTH];
    memset(hostname, 0, sizeof (hostname));
    memset(&simStatus, 0, sizeof (Status_t));

    int i = 0;

    FILE *dataFile = fopen(DATAFILENAME, "rt");
    
    char line[MAXDATAFILELINELENGTH];
    const char delimiters[] = "\", ";
    char *token;

    fgets(line, sizeof (line), dataFile); //removing header;

    while (fgets(line, sizeof (line), dataFile) && i < MAXSIMULATIONSTEPS) {
        token = strtok(line, delimiters);
        position.Longitude = atof(token);
        token = strtok(NULL, delimiters);
        position.Latitude = atof(token);
        token = strtok(NULL, delimiters);
        position.Altitude = atof(token);
        token = strtok(NULL, delimiters);
        position.Status = atoi(token);

        token = strtok(NULL, delimiters);
        notification.Height = atof(token);
        token = strtok(NULL, delimiters);
        notification.Battery = atoi(token);
        token = strtok(NULL, delimiters);
        notification.Heading = atof(token);
        token = strtok(NULL, delimiters);
        notification.Speed = atof(token);

        token = strtok(NULL, delimiters);
        simStatus.Scenario = atoi(token);
        token = strtok(NULL, delimiters);
        simStatus.State = atoi(token);
        token = strtok(NULL, delimiters);
        simStatus.Positioning = atoi(token);
        token = strtok(NULL, delimiters);
        simStatus.Searching = atoi(token);

        token = strtok(NULL, delimiters);
        posCl1.Longitude = atof(token);
        token = strtok(NULL, delimiters);
        posCl1.Latitude = atof(token);
        token = strtok(NULL, delimiters);
        posCl1.Altitude = atof(token);
        token = strtok(NULL, delimiters);
        posCl1.Status = atoi(token);
        token = strtok(NULL, delimiters);
        client1.MAC[0] = 1;
        client1.MAC[1] = 2;
        client1.MAC[2] = 3;
        client1.MAC[3] = 4;
        client1.MAC[4] = 5;
        client1.MAC[5] = 6;

        token = strtok(NULL, delimiters);
        posCl2.Longitude = atof(token);
        token = strtok(NULL, delimiters);
        posCl2.Latitude = atof(token);
        token = strtok(NULL, delimiters);
        posCl2.Altitude = atof(token);
        token = strtok(NULL, delimiters);
        posCl2.Status = atoi(token);
        token = strtok(NULL, delimiters);
        client2.MAC[0] = 6;
        client2.MAC[1] = 5;
        client2.MAC[2] = 4;
        client2.MAC[3] = 3;
        client2.MAC[4] = 2;
        client2.MAC[5] = 1;

        token = strtok(NULL, delimiters);
        gethostname(hostname, MAXHOSTNAMELENGTH);
	strcpy(simStatus.Hostname, hostname); //using real hostname, not the one from simulationdata.csv

        client1.Position = posCl1;
        client2.Position = posCl2;

        simStatus.Client1 = client1;
        simStatus.Client2 = client2;

        notification.Position = position;
	simStatus.Position = position;
        notification.Status = simStatus;

        pthread_mutex_lock(&currentStatusMutex);
        currentStatus = simStatus;
        pthread_mutex_unlock(&currentStatusMutex);
	
        notifications[i] = notification;
        i++;
    }
    
    simulationSteps = i;
}

void getSimulatedNotificationValues(Notification_t * notification) {
    extern int simulationStep;
    if (simulationStep < simulationSteps) {
        *notification = notifications[simulationStep];
        simulationStep++;
    } else {
        //restarting simulation
        simulationStep = 0;
    }
}
