#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <ifaddrs.h>


#include "types.h"
#include "uavext.h"
#include "constants.h"
#include "socket.h"
#include "gps.h"

#include "../common_files/types.h"
#include "../common_files/helper.h"

struct addrinfo hints_notification_send, *servinfo_notification_send;
struct addrinfo hints_ping_send, *servinfo_ping_send;

int generateControllerCommunicationSocket() {

    int sockfd;
    struct addrinfo hints, *servinfo, *p;
    struct sigaction sa;
    int yes = 1;
    int rv;
    int port = UAV_PORT_COMMUNICATION;
    char charPort[6];
    memset(charPort, 0, sizeof charPort);
    sprintf(charPort, "%d", port);


    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE; // use my IP

    if ((rv = getaddrinfo(NULL, charPort, &hints, &servinfo)) != 0) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
        return 1;
    }

    // loop through all results and bind to the first
    for (p = servinfo; p != NULL; p = p->ai_next) {
        if ((sockfd = socket(p->ai_family, p->ai_socktype,
                p->ai_protocol)) == -1) {
            printf("error: server: socket\n");
            continue;
        }

        if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes,
                sizeof (int)) == -1) {
            printf("error: setsockopt\n");
            exit(1);
        }

        if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
            close(sockfd);
            perror("server: bind");
            continue;
        }

        break;
    }

    if (p == NULL) {
        fprintf(stderr, "server: failed to bind\n");
        return 2;
    }

    freeaddrinfo(servinfo);
    
    if (listen(sockfd, MAXPENDINGCONNECTIONS) == -1) {
        perror("listen");
        exit(1);
    }

    sa.sa_handler = sigchld_handler; // reap all dead processes
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;
    if (sigaction(SIGCHLD, &sa, NULL) == -1) {
        perror("sigaction");
        exit(1);
    }

    return sockfd;

}

int generateClientCommunicationSocket(char *hostname) {

    int sockfd;
    struct addrinfo hints, *servinfo, *p;
    int rv;
    char s[INET6_ADDRSTRLEN];
    int port = UAV_PORT_COMMUNICATION;
    char charPort[6];
    memset(charPort, 0, sizeof charPort);
    sprintf(charPort, "%d", port);

    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    if ((rv = getaddrinfo(hostname, charPort, &hints, &servinfo)) != 0) {
        printf("error: getaddrinfo: %s\n", gai_strerror(rv));
        return 1;
    }

    for (p = servinfo; p != NULL; p = p->ai_next) {
        if ((sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
            printf("error: client: socket");
            continue;
        } 
        if (connect(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
            close(sockfd);
            perror("client: connect");
            continue;
        } 
        break;
    }

    if (p == NULL) {
        fprintf(stderr, "client: failed to connect\n");
        return 2;
    }
    inet_ntop(p->ai_family, get_in_addr((struct sockaddr *) p->ai_addr), s, sizeof s);

    freeaddrinfo(servinfo);
    return sockfd;

}

void correctHost(char * host) {
    char * pos;
    pos = strchr(host, '%');
    if(pos == NULL) {
    } else {
        int i = 0;
        for(i = 0; i < strlen(pos); i++) {
            *(pos + i) = '\0';
        }
    }
}

int generateControllerPingSocket(int ifn) {
    int sockfd_udp;
    int status;
    int port = UAV_PORT_PING;
    char *multicastaddr = MULTICASTADDR;

    char charPort[6];
    memset(charPort, 0, sizeof charPort);
    sprintf(charPort, "%d", port);

    memset(&hints_ping_send, 0, sizeof (hints_ping_send));
    hints_ping_send.ai_family = AF_UNSPEC; //no broadcast possible with ipv6 -> would need multicast stuff
    hints_ping_send.ai_socktype = SOCK_DGRAM;

    if (status = getaddrinfo(multicastaddr, charPort, &hints_ping_send, &servinfo_ping_send) != 0) {
        fprintf(stderr, "getaddrinfo ping socket error: %s\n", gai_strerror(status));
        exit(1);
    }

    sockfd_udp = createSocket(servinfo_ping_send->ai_family, servinfo_ping_send->ai_socktype, servinfo_ping_send->ai_protocol);

    if (servinfo_ping_send->ai_family == AF_INET) {
        printf("sending ping: ipv4\n");
        if (setsockopt(sockfd_udp, SOL_SOCKET, SO_BROADCAST, &ifn, sizeof (ifn)) < 0) {
            printf("setsockopt ipv4 error\n");
            return EXIT_FAILURE;
        }
    } else if (servinfo_ping_send->ai_family == AF_INET6) {
        printf("sending ping: ipv6\n");
        if (setsockopt(sockfd_udp, IPPROTO_IPV6, IPV6_MULTICAST_IF, &ifn, sizeof (ifn)) < 0) {
            printf("setsockopt ipv6 error\n");
            return EXIT_FAILURE;
        }
    }

    return sockfd_udp;
}

int generateControllerNotificationSocket(int ifn, struct sockaddr_storage subscriber, char *port) {
    int sockfd_udp;
    char host[NI_MAXHOST];
    int s;
    int family = subscriber.ss_family;
    
    family = AF_INET;
    subscriber.ss_family = family;
    

    memset(&hints_notification_send, 0, sizeof (hints_notification_send));

    hints_notification_send.ai_family = family;
    hints_notification_send.ai_socktype = SOCK_DGRAM;

    s = getnameinfo(getCastedSockaddr(&subscriber), (family == AF_INET) ? sizeof (struct sockaddr_in) : sizeof (struct sockaddr_in6), host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);
    if (s != 0) {
        printf("getnameinfo() uavext generateSocket failed: %s\n", gai_strerror(s));
        exit(EXIT_FAILURE);
    }

    int status;

    if (status = getaddrinfo(host, port, &hints_notification_send, &servinfo_notification_send) != 0) {
        fprintf(stderr, "getaddrinfo error: %s\n", gai_strerror(status));
        exit(1);
    }
    struct addrinfo *p;

    for (p = servinfo_notification_send; p != NULL; p = p->ai_next) {
         if((sockfd_udp = createSocket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
            printf("error ping listener: wrong socket\n");
            continue;
        }
         break;
    }
    return sockfd_udp;
}

int generateUdpPingListeningSocket(int port, char* iface) {
    int domain = AF_UNSPEC;
    int type = SOCK_DGRAM;
    int socketNbr, rv;
    struct addrinfo hints, *servinfo, *p;
    char charPort[6];
    memset(charPort, 0, sizeof charPort);
    sprintf(charPort, "%d", port);
    char *multicastaddr = MULTICASTADDR;
    int ifn = 0;

    memset(&hints, 0, sizeof (hints));
    hints.ai_family = domain;
    hints.ai_socktype = type;
    hints.ai_protocol = IPPROTO_UDP;
    hints.ai_flags = AI_PASSIVE | AI_NUMERICHOST; // use my IP

    if ((rv = getaddrinfo(multicastaddr, charPort, &hints, &servinfo)) != 0) {
        printf("error ping listener socket  getaddrinfo: %s\n", gai_strerror(rv));
        return FAILURE;
    }
    ifn = getIfn((char*)iface);

    ((struct sockaddr_in6 *)servinfo->ai_addr)->sin6_scope_id = ifn;

    p = servinfo;
    for (p = servinfo; p != NULL; p = p->ai_next) {
        if ((socketNbr = createSocket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
            printf("error ping listener: wrong socket\n");
            continue;
        }

        int n = 0;
        if (bind(socketNbr, p->ai_addr, p->ai_addrlen) == -1) {
            close(socketNbr);
            printf("error ping listener: bind: %s\n", gai_strerror(n));
            continue;
        }
        break;
    }
    if (p == NULL) {
        printf("error ping listener: failed to bind socket\n");
        freeaddrinfo(servinfo);
    }

    if(p->ai_family == AF_INET6) {
        struct ipv6_mreq joinRequest;
        if(inet_pton(p->ai_family, multicastaddr, &joinRequest.ipv6mr_multiaddr) == 0) {
            printf("bad address %s\n", multicastaddr);
        }
        char str[INET6_ADDRSTRLEN];
        inet_ntop(AF_INET6, &((struct sockaddr_in6 *)p->ai_addr)->sin6_addr, str, INET6_ADDRSTRLEN);
        joinRequest.ipv6mr_interface = if_nametoindex("eth1");
        if(setsockopt(socketNbr, IPPROTO_IPV6, IPV6_JOIN_GROUP, &joinRequest, sizeof(joinRequest)) < 0) {
            printf("setsockopt IPV6_JOIN_GROUP failed\n");
        }
    } else if(p->ai_family == AF_INET) {
        struct ip_mreq joinRequest;
        joinRequest.imr_multiaddr = ((struct sockaddr_in *) p->ai_addr)->sin_addr;
        joinRequest.imr_interface.s_addr = 0;
        if(setsockopt(socketNbr, IPPROTO_IP, IP_ADD_MEMBERSHIP, &joinRequest, sizeof(joinRequest)) < 0) {
            printf("setsockopt IP_ADD_MEMBERSHIP failed\n");
        }
    } else {
        printf("unknown address family");
        exit(EXIT_FAILURE);
    }

    freeaddrinfo(servinfo);
    return socketNbr;
}

int generateUdpNotificationListeningSocket(int port) {
    int domain = AF_INET6;
    int type = SOCK_DGRAM;
    int socketNbr, rv;
    struct addrinfo hints, *servinfo, *p;
    char charPort[6];
    memset(charPort, 0, sizeof charPort);
    sprintf(charPort, "%d", port);

    memset(&hints, 0, sizeof (hints));
    hints.ai_family = domain;
    hints.ai_socktype = type;
    hints.ai_flags = AI_PASSIVE; // use my IP

    if ((rv = getaddrinfo(NULL, charPort, &hints, &servinfo)) != 0) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
        return FAILURE;
    }

    for (p = servinfo; p != NULL; p = p->ai_next) {
        if ((socketNbr = createSocket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
            printf("notificationlistener: socket");
            continue;
        }

        if (bind(socketNbr, p->ai_addr, p->ai_addrlen) == -1) {
            close(socketNbr);
            printf("notificationlistener: bind");
            continue;
        }

        break;
    }
    if(p->ai_family == AF_INET6) {
        printf("ipv6\n");
    } else     if(p->ai_family == AF_INET) {
        printf("ipv4\n");
    }
    if (p == NULL) {
        fprintf(stderr, "notificationlistener: failed to bind socket\n");
        return FAILURE;
    }

    freeaddrinfo(servinfo);
    return socketNbr;
}

int sendPing(int sockfd_udp, char* output) {
    int sentBytes = 0;
    sentBytes = sendto(sockfd_udp, output, strlen(output), 0, servinfo_ping_send->ai_addr, servinfo_ping_send->ai_addrlen);
    freeaddrinfo(servinfo_ping_send);

    return sentBytes;
}

int sendNotification(int sockfd_udp, char* output) {
    int sentBytes = 0;
    char receiver[INET6_ADDRSTRLEN];
    int tmpTestSocket;

    inet_ntop(servinfo_notification_send->ai_family, get_in_addr(servinfo_notification_send->ai_addr), receiver, sizeof (receiver));
    printf("notified %s\n", receiver);

    struct addrinfo *p;

    for (p = servinfo_notification_send; p != NULL; p = p->ai_next) {
        if ((tmpTestSocket = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
            perror("talker: socket");
            continue;
        }
        close(tmpTestSocket);
        break;
    }

    if (p == NULL) {
        fprintf(stderr, "talker: failed to bind socket\n");
        return 2;
    }


    if ((sentBytes = sendto(sockfd_udp, output, strlen(output), 0, p->ai_addr, p->ai_addrlen)) == -1) {
        perror("talker: sendto");
	printf("exiting...\n");
        exit(1);
    }
    inet_ntop(p->ai_addr->sa_family, get_in_addr(p->ai_addr), receiver, sizeof (receiver));
    freeaddrinfo(servinfo_notification_send);

    return sentBytes;
}

int sendTcpPacketOnExistingSocket(int sockfd, char * output, int outputLength, int expectedResponse, int retriesOnFailure) {
    int sentBytes = 0;
    int answer = FAILURE;
    int tries = 1 + retriesOnFailure;
    int currentSendTry = 0;

    for (currentSendTry = 0; currentSendTry < tries; currentSendTry++) {
        sentBytes = send(sockfd, output, outputLength, 0);
        if (sentBytes < 0) {
            printf("ERROR writing to socket\n");
            answer = FAILURE;
        } else {
            if (expectedResponse != FALSE) {
                answer = readSocketInput(sockfd);
                printf("answer: %d\n", answer);
                if (answer == FALSE) {
                    return -1;
                }
                if (answer == expectedResponse) {
                    break;
                }
            } else {
            }
        }
    }
    return answer;
}

int sendTcpPacketOnNewSocket(char * hostname, char * output, int outputLength, int expectedResponse, int retriesOnFailure) {
    int result;
    int sockfd = generateClientCommunicationSocket(hostname);
    result = sendTcpPacketOnExistingSocket(sockfd, output, outputLength, expectedResponse, retriesOnFailure);
    closeSocket(sockfd);
    return result;
}

int receiveTcpPacket(int sockfd, char * inbuffer, int bufferlength) {
    int numBytes;

    numBytes = recv(sockfd, inbuffer, bufferlength, 0);
    return numBytes;
}

int receiveUdpPacket(int socket, char *buf, char *sender, int sendersize) {
    int numbytes = 0;
    struct sockaddr_storage their_addr;
    socklen_t addr_len;

    addr_len = sizeof (their_addr);
    numbytes = recvfrom(socket, buf, MAXBUFLEN - 1, 0, (struct sockaddr *) & their_addr, &addr_len);
    inet_ntop(their_addr.ss_family, &(((struct sockaddr_in *) & their_addr)->sin_addr), sender, sendersize);

    return numbytes;
}

int acceptControllerSocket(int sockfd_tcp) {

    struct sockaddr_storage their_addr; // connector's address information
    socklen_t sin_size;
    int accepted_sockfd;
    char s[INET6_ADDRSTRLEN];

    sin_size = sizeof their_addr;
    accepted_sockfd = accept(sockfd_tcp, (struct sockaddr *) & their_addr, &sin_size);
    if (accepted_sockfd == -1) {
        perror("accept");
        return FAILURE;
    }

    inet_ntop(their_addr.ss_family, get_in_addr((struct sockaddr *) & their_addr), s, sizeof s);
    printf("server: got new connection from %s on socket %d\n", s, sockfd_tcp);

    return accepted_sockfd;
}

void closeSocket(int sockfd_tcp) {
    close(sockfd_tcp);
}

int readSocketInput(int sockfd) {
    int n = 0;
    int i = 0;
    int inputStringCounter = 0;
    int result = 0;
    int code = 0;
    char inputString[MAXLINELENGTH];
    char input[MAXLINELENGTH];
    char answer[MAXLINELENGTH];
    char In1;
    unsigned char incommand;
    memset(inputString, '\0', sizeof ( inputString));
    memset(input, '\0', sizeof ( input));
    memset(answer, '\0', sizeof ( answer));

    n = receiveTcpPacket(sockfd, input, MAXDATALENGTH);
    if (n < 0) printf("ERROR reading from socket\n");
    if (n == 0) {
        printf("ERROR, remote host closed the connection\n");
        return FALSE;
    }
    if (n > 0) {
        resetInputString(inputString, sizeof (inputString), &inputStringCounter);

        for (i = 0; i < n; i++) //for all chars in string
        {
            In1 = input[i];
            inputString[inputStringCounter++] = In1;
            if (In1 == '\r') {
                result = socketCommandReceived(inputString, strlen(inputString), &incommand, answer);
                resetInputString(inputString, sizeof ( inputString), &inputStringCounter);
                switch (incommand) {
                    case 'r': // TRUE or FALSE
                        code = (int) * answer;
                        return code;
                        break;
                    case 'd': //flight direction request
                        return incommand;
                        break;
                }
            }
            if (inputStringCounter >= sizeof ( inputString) - 1) {
                resetInputString(inputString, sizeof ( inputString), &inputStringCounter);
            }
        }
    }
    return FAILURE;
}

int socketCommandReceived(char *inputString, int length, uint8_t* command, char* data) {
    if (strlen(inputString) >= 5 && checkCRC(inputString, strlen(inputString) - 1) == SUCCESS) {
        decodeExternalCommand(inputString, command, data);
        return SUCCESS;

    } else { //too short string or wrong crc
        return FAILURE;
    }
}

void inSocketCommand(char *inputString, int length, uint8_t* command, char* data) {
    int i = 0;
    if (strlen(inputString) > 5 && checkCRC(inputString, strlen(inputString) - 1) == SUCCESS) {
        decodeExternalCommand(inputString, command, data);
    } else { //too short string or wrong crc or no real command
        printf("wrong crc received in: %s\r\n", inputString);
        for (i = 0; i < strlen(inputString); i++) {
            printf("%c", inputString[i]);
        }
        printf("\r\n");
    }
}

void *get_in_addr(struct sockaddr *sa) {
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*) sa)->sin_addr);
    }
    return &(((struct sockaddr_in6*) sa)->sin6_addr);
}

void *getCastedSockaddr(struct sockaddr_storage *sa) {
    if (sa->ss_family == AF_INET) {
        return ((struct sockaddr_in*) sa);
    }
    return ((struct sockaddr_in6*) sa);
}

void sigchld_handler(int s) {
    while (waitpid(-1, NULL, WNOHANG) > 0);
}
