#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include "serial.h"

#define MODEMDEVICE             "/dev/ttyUSB0"
#define _POSIX_SOURCE 1         //POSIX compliant source

long BAUDRATE = B57600;
long DATABITS = CS8;
long STOPBITS = 0;
long PARITYON = 0;
long PARITY = 0;
int Data_Bits = 8;              // Number of data bits
int Stop_Bits = 1;              // Number of stop bits
int Parity = 0;                 // Parity as follows:

void openSerialPort ( char *devicename , int *fd, FILE **input, FILE **output, int *tty, struct termios *oldkey, struct termios *newkey, struct sigaction *saio, struct termios *oldtio, struct termios *newtio) {
    openCOMPort ( devicename, fd );
    configPort(fd, oldtio, newtio);
}


void openCOMPort ( char *devicename, int *fd ) {
     *fd = open ( devicename, O_RDWR | O_NOCTTY | O_NONBLOCK | O_NDELAY);
    if ( *fd < 0 )
    {
        printf ( "error: could not open %s\n", devicename );
    }
}


void configSerialHandler(struct sigaction *saio) {
    saio->sa_handler = signal_handler_IO;
    sigemptyset ( &saio->sa_mask );   //saio.sa_mask = 0;
    saio->sa_flags = 0;
    sigaction ( SIGIO,saio,NULL );
}

void configPort( int *fd , struct termios *oldtio, struct termios *newtio) {
    tcgetattr ( *fd,oldtio ); 	// save current port settings
    newtio->c_cflag = BAUDRATE | DATABITS | STOPBITS | PARITYON | PARITY | CLOCAL | CREAD;
    newtio->c_iflag = IGNPAR | ICRNL; //ICRNL   : map CR to NL (otherwise a CR input on the other computer will not terminate input)
    newtio->c_oflag = 0;	// raw output
    newtio->c_lflag = 0;        //ICANON;
    newtio->c_cc[VMIN]=0;
    newtio->c_cc[VTIME]=0;
    tcflush ( *fd, TCIFLUSH );
    tcsetattr ( *fd,TCSANOW,newtio );
}

void openTerminals(FILE **input, FILE **output) {
    *input = fopen ( "/dev/tty", "r" );   //open the terminal keyboard
    *output = fopen ( "/dev/tty", "w" );  //open the terminal screen
	
    if ( !*input || !*output ) {
        printf ( "Unable to open /dev/tty\n" );
        exit ( 1 );
    }
}

void configTTY(int *tty, struct termios *oldkey, struct termios *newkey) {
    *tty = open ( "/dev/tty", O_RDWR | O_NOCTTY | O_NONBLOCK ); //set the user console port up
    tcgetattr ( *tty,oldkey ); // save current port settings   //so commands are interpreted right for this program
    newkey->c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
    newkey->c_iflag = IGNPAR;
    newkey->c_oflag = 0;
    newkey->c_lflag = 0;       //ICANON;
    newkey->c_cc[VMIN]=1;
    newkey->c_cc[VTIME]=0;
    tcflush ( *tty, TCIFLUSH );
    tcsetattr ( *tty,TCSANOW, newkey );
}

void closeSerialPort(int *fd , int *tty, FILE **input, FILE **output, struct termios *oldtio, struct termios *oldkey) {
    tcsetattr ( *fd,TCSANOW,oldtio );
    close ( *fd );
}
