#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "types.h"
#include "uavint.h"
#include "constants.h"

#include "../common_files/types.h"
#include "../common_files/helper.h"


int serialInputAvailable = FALSE; //TRUE while no signal received
int serialTimeout = 10; //seconds

int inSerialCommand(char *inputString, int length, char expectedCommand, int sockfd) {
    int i = 0;
    uint16_t address;
    uint8_t command;
    char data[200];
    char debugString[length];
    memset(debugString, '\0', sizeof ( debugString));
    memset(data, '\0', sizeof ( data));
    strncpy(debugString, inputString, length - 1);
    if (strlen(inputString) > 5 && checkCRC(inputString, strlen(inputString) - 1) == SUCCESS) {
        decodeInternalCommand(inputString, &address, &command, data);
        if (command != expectedCommand) {
            return FAILURE;
        }
        return SUCCESS;
    } else { //too short string or wrong crc
        return FAILURE;
    }
}


int readSerialInput(char expectedCommand, int fd, int sockfd) {

    extern volatile int listeningForSerialInput;
    char In1;
    int res;
    int i;
    char inputString[200];
    char buf[200];
    int inputStringCounter;
    int result = FAILURE;

    memset(buf, '\0', sizeof ( buf));
    memset(inputString, '\0', sizeof ( inputString));

    resetInputString(inputString, sizeof ( inputString), &inputStringCounter);

    time_t starttime;
    time_t now;
    time(&starttime);
    time(&now);

    while (listeningForSerialInput == TRUE) {

        time(&now);
        if (((int) (difftime(now, starttime))) > serialTimeout) {
            result = FAILURE;
            break;
        }
        if (serialInputAvailable == TRUE) //if input is available
        {
            res = 0;
            memset(buf, '\0', sizeof ( buf));
            res = read(fd, buf, sizeof ( buf));
            if (res > 0) {
                for (i = 0; i < res; i++) //for all chars in string
                {
                    In1 = buf[i];
                    inputString[inputStringCounter++] = In1;
                    if (In1 == '\r') {
                        if (inSerialCommand(inputString, strlen(inputString), expectedCommand, sockfd) == SUCCESS) {
                            printf("command received: %s\n", inputString);
                            result = SUCCESS;
                        }
                        listeningForSerialInput = FALSE;
                        resetInputString(inputString, sizeof ( inputString), &inputStringCounter);

                    }
                    if (inputStringCounter >= sizeof ( buf) - 1) {
                        resetInputString(inputString, sizeof ( inputString), &inputStringCounter);
                    }
                } //end of for all chars in string
            } //end if res>0
            serialInputAvailable = FALSE; /* wait for new input */
        } //end if wait flag == FALSE
    }
    return result;
}

void signal_handler_IO(int status) {
    serialInputAvailable = TRUE;
}

void setListeningForSerialInput(int status) {
    extern volatile int listeningForSerialInput;
    listeningForSerialInput = status;
}
